#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# the event handling bits for construction lines
# and circles
#

import math

import gtk

import Generic.point as point
import Generic.segment as segment
import Generic.circle as circle
import Generic.arc as arc
import Generic.hcline as hcline
import Generic.vcline as vcline
import Generic.acline as acline
import Generic.cline as cline
import Generic.ccircle as ccircle
import Generic.util as util


#
# horizontal construction lines
#

def hcline_create(gtkimage, tool):
    assert len(tool) == 1, "HCLine should hold a single object."
    _p = tool.popObject()
    _active_layer = gtkimage.getActiveLayer()
    if _active_layer.findObject(_p) is None:
        _active_layer.addObject(_p)
    gtkimage.addObject(hcline.HCLine(_p))
    gtkimage.redraw()
    hcline_mode_init(tool)
    
def hcline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = point.Point(_x, _y)
    tool.pushObject(_pt)
    hcline_create(gtkimage, tool)

def hcline_entry_event_cb(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            val = eval(text)
        except:
            print "Invalid value: " + `text` # need better error handling
            return
        _val = val
        if not isinstance(_val, float):
            try:
                _val = float(val)
            except:
                print "Invalid value: " + `val` # need better error handling
                return
        p, flag = gtkimage.findPoint(0.0, _val)
        if p is None:
            p = point.Point(0.0, _val)
        tool.pushObject(p)
        hcline_create(gtkimage, tool)
            
def hcline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", hcline_button_press_cb)
    tool.setHandler("entry_event", hcline_entry_event_cb)

#
# vertical construction lines
#

def vcline_create(gtkimage, tool):
    assert len(tool) == 1, "VCLine should hold a single object."
    _p = tool.popObject()
    _active_layer = gtkimage.getActiveLayer()
    if _active_layer.findObject(_p) is None:
        _active_layer.addObject(_p)
    gtkimage.addObject(vcline.VCLine(_p))
    gtkimage.redraw()
    vcline_mode_init(tool)
    
def vcline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = point.Point(_x, _y)
    tool.pushObject(_pt)
    vcline_create(gtkimage, tool)

def vcline_entry_event_cb(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            val = eval(text)
        except:
            print "Invalid value: " + `text` # need better error handling
            return
        _val = val
        if not isinstance(_val, float):
            try:
                _val = float(val)
            except:
                print "Invalid value: " + `val` # need better error handling
                return
        p, flag = gtkimage.findPoint(_val, 0.0)
        if p is None:
            p = point.Point(_val, 0.0)
        tool.pushObject(p)
        vcline_create(gtkimage, tool)

def vcline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", vcline_button_press_cb)
    tool.setHandler("entry_event", vcline_entry_event_cb)

#
# common functions for for acline, cline, and ccircle objects
#

def conline_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _width, _height = gtkimage.getSize()
    _gc = gtkimage.getGC()
    _x = event.x
    _y = event.y
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        widget.window.draw_line(_gc, _px, _py, _xc, _yc)
    tool.setCurrentPoint(_x, _y)
    widget.window.draw_line(_gc, _px, _py, _x, _y)

def make_tuple(text):
    try:
        _tpl = eval(text)
    except:
        raise StandardError, "Invalid point: " + `text`
    if not isinstance(_tpl, tuple):
        raise TypeError, "Invalid point: " + `_tpl`
    if len(_tpl) != 2:
        raise ValueError, "Invalid point: " + `_tpl`
    return _tpl
    
#
# angled construction lines
#

def acline_create(gtkimage, tool):
    assert len(tool) == 2, "ACLine tool should hold two objects."
    _obj = tool.popObject() # a float or a Point
    _loc = tool.popObject() # a Point
    _x, _y = _loc.getCoords()
    if isinstance(_obj, float):
        _angle = _obj
    elif isinstance(_obj, point._Point):
        _ox, _oy = _obj.getCoords()
        _a = (180.0/math.pi)*math.atan2((_oy - _y), (_ox - _x))
        try:
            _angle = util.make_angle(_a)
        except:
            print "Invalid angle" # need better error handling ...
            return
    else:
        raise "Unexpected object type: " + `_obj`
    _active_layer = gtkimage.getActiveLayer()
    if _active_layer.findObject(_loc) is None:
        _active_layer.addObject(_loc)
    gtkimage.addObject(acline.ACLine(_loc, _angle))
    gtkimage.redraw()
    gtkimage.setPrompt("Click in drawing area or enter a Point")
    acline_mode_init(tool)
        
def acline_entry_make_angle(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            angle = util.make_angle(text)
        except:
            print "Invalid angle: " + `text`
            return
        tool.pushObject(angle)
        acline_create(gtkimage, tool)

def acline_entry_make_pt(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            tpl = make_tuple(text)
        except:
            print "Invalid point: " + `text` # need better error handling
            return
        x, y = tpl
        p, flag = gtkimage.findPoint(x, y)
        if p is None:
            p = point.Point(x, y)
        tool.pushObject(p)
        if len(tool) == 1:
            _x, _y = p.getCoords()
            tool.storeCoords(_x, _y)
            tool.setLocation(_x, _y)
            tool.setHandler("entry_event", acline_entry_make_angle)
            tool.setHandler("motion_notify", conline_motion_notify_cb)
            gtkimage.setPrompt("Enter the angle or click in the drawing area")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)
        else:
            acline_create(gtkimage, tool)
        
def acline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = point.Point(_x, _y)
    tool.pushObject(_pt)
    if len(tool) == 1:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("entry_event", acline_entry_make_angle)
        tool.setHandler("motion_notify", conline_motion_notify_cb)
        gtkimage.setPrompt("Enter the angle or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
    else:
        acline_create(gtkimage, tool)

def acline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", acline_button_press_cb)
    tool.setHandler("entry_event", acline_entry_make_pt)

#
# two point construction line
#

def cline_create(gtkimage, tool):
    assert len(tool) == 2, "CLine tool should hold two objects."
    _active_layer = gtkimage.getActiveLayer()
    _p2 = tool.popObject()
    if _active_layer.findObject(_p2) is None:
        _active_layer.addObject(_p2)
    _p1 = tool.popObject()
    if _active_layer.findObject(_p1) is None:
        _active_layer.addObject(_p1)
    gtkimage.addObject(cline.CLine(_p1, _p2))
    gtkimage.redraw()
    gtkimage.setPrompt("Click in drawing area or enter a Point")
    cline_mode_init(tool)

def cline_entry_make_pt(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            tpl = make_tuple(text)
        except:
            print "Invalid point: " + `text` # need better error handling
            return
        x, y = tpl
        p, flag = gtkimage.findPoint(x, y)
        if p is None:
            p = point.Point(x, y)
        tool.pushObject(p)
        if len(tool) == 2:
            cline_create(gtkimage, tool)
        else:
            _x, _y = p.getCoords()
            tool.storeCoords(_x, _y)
            tool.setLocation(_x, _y)
            tool.setHandler("motion_notify", conline_motion_notify_cb)
            gtkimage.setPrompt("Enter the second Point or click in the drawing area")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)

def cline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = point.Point(_x, _y)
    tool.pushObject(_pt)
    if len(tool) == 1:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", conline_motion_notify_cb)
        gtkimage.setPrompt("Enter the second point or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
    else:
        cline_create(gtkimage, tool)

def cline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", cline_button_press_cb)
    tool.setHandler("entry_event", cline_entry_make_pt)

#
# construction circles
#

def ccircle_make_point(gtkimage, text, addflag=True):
    _tpl = make_tuple(text)
    _x, _y = _tpl
    _pt = gtkimage.findPoint(_x, _y)
    if _pt is None:
        _pt = point.Point(_x, _y)
    return _pt

def ccircle_make_radius(text):
    _text_rad = eval(text)
    _rad = _text_rad
    if not isinstance(_rad, float):
        _rad = float(_text_rad)
    if _rad < 1e-10:
        raise ValueError, "Invalid radius: " + `_rad`
    return _rad

def ccircle_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _width, _height = gtkimage.getSize()
    _gc = gtkimage.getGC()
    _x = event.x
    _y = event.y
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _radius = math.hypot((_px - _xc),(_py - _yc))
        _xmin = _px - _radius
        _ymin = _py - _radius
        _cw = _ch = _radius * 2
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch,
                               0, 360*64)
    tool.setCurrentPoint(_x, _y)
    _radius = math.hypot((_px - _x),(_py - _y))
    _xmin = _px - _radius
    _ymin = _py - _radius
    _cw = _ch = _radius * 2
    widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch, 0, 360*64)

def ccircle_entry_make_radius(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    if len(text):
        try:
            radius =  ccircle_make_radius(text)
        except:
            print "Invalid radius: " + `text` # need better error handling
            return
        tool.pushObject(radius)
        ccircle_create(gtkimage, tool)

def ccircle_create(gtkimage, tool):
    assert len(tool) == 2, "CCircle tool should hold two objects."
    _obj = tool.popObject() # radius or a Point
    _cen = tool.popObject() # center Point
    if isinstance(_obj, float):
        _radius = _obj
    elif isinstance(_obj, point._Point):
        _radius = _obj - _cen
    else:
        raise "Unexpected object type: " + `_obj`
    _active_layer = gtkimage.getActiveLayer()
    if _active_layer.findObject(_cen) is None:
        _active_layer.addObject(_cen)
    gtkimage.addObject(ccircle.CCircle(_cen, _radius))
    gtkimage.redraw()
    gtkimage.setPrompt("Click in drawing area or enter a Point")
    init_func = tool.getHandler("initialize")
    init_func(tool)

def ccircle_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        try:
            _p = ccircle_make_point(gtkimage, _text)
        except:
            print "Invalid point: " + `_text` # need better error handling
            return
        tool.pushObject(_p)
        if len(tool) == 1:
            _x, _y = _p.getCoords()
            tool.storeCoords(_x, _y)
            tool.setLocation(_x, _y)
            tool.setHandler("motion_notify", ccircle_motion_notify_cb)
            tool.setHandler("entry_event", ccircle_entry_make_radius)
            gtkimage.setPrompt("Enter the radius or click in the drawing area")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)
        else:
            ccircle_create(gtkimage, tool)

def ccircle_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = point.Point(_x, _y)
    tool.pushObject(_pt)
    if len(tool) == 1:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", ccircle_motion_notify_cb)
        tool.setHandler("entry_event", ccircle_entry_make_radius)
        gtkimage.setPrompt("Enter the radius or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
    else:
        ccircle_create(gtkimage, tool)

def ccircle_cpmode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", ccircle_button_press_cb)
    tool.setHandler("entry_event", ccircle_entry_make_pt)
    tool.setHandler("initialize", ccircle_cpmode_init)

#
# two-point construction circle
#

def ccircle_tp_create(gtkimage, tool):
    assert len(tool) == 2, "CCircle TP tool should hold two objects."
    _p2 = tool.popObject()
    _p1 = tool.popObject()
    _xc = (_p1.x + _p2.x)/2.0
    _yc = (_p1.y + _p2.y)/2.0
    _rad = (_p2 - _p1)/2.0
    _pt, _new_pt = gtkimage.findPoint(_xc, _yc)
    if _pt is None:
        _pt = point.Point(_xc, _yc)
        gtkimage.addObject(_pt)
    gtkimage.addObject(ccircle.CCircle(_pt, _rad))
    gtkimage.redraw()
    gtkimage.setPrompt("Click in drawing area or enter a Point")
    init_func = tool.getHandler("initialize")
    init_func(tool)


def ccircle_tp_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getCoordPair(0)
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _gc = gtkimage.getGC()
    _x = event.x
    _y = event.y
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _xcen = _px + (_xc - _px)//2
        _ycen = _py + (_yc - _py)//2
        _radius = math.hypot((_xc - _xcen), (_yc - _ycen))
        _xmin = _xcen - _radius
        _ymin = _ycen - _radius
        _cw = _ch = _radius * 2
        widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch,
                               0, 360*64)
    tool.setCurrentPoint(_x, _y)
    _xcen = _px + (_x - _px)//2
    _ycen = _py + (_y - _py)//2
    _radius = math.hypot((_x - _xcen), (_y - _ycen))
    _xmin = _xcen - _radius
    _ymin = _ycen - _radius
    _cw = _ch = _radius * 2
    widget.window.draw_arc(_gc, gtk.FALSE, _xmin, _ymin, _cw, _ch, 0, 360*64)

def ccircle_tp_entry_make_pt(gtkimage, widget, tool):
    _entry = gtkimage.getEntry()
    _text = _entry.get_chars(0,-1)
    _entry.delete_text(0,-1)
    if len(_text):
        try:
            _p = ccircle_make_point(gtkimage, _text, False)
        except:
            print "Invalid point: " + `_text` # need better error handling
            return
        tool.pushObject(_p)
        if len(tool) == 1:
            _x, _y = _p.getCoords()
            tool.storeCoords(_x, _y)
            tool.setLocation(_x, _y)
            tool.setHandler("motion_notify", ccircle_tp_motion_notify_cb)
            gtkimage.setPrompt("Enter the second point or click in the drawing area")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)
        else:
            ccircle_tp_create(gtkimage, tool)

def ccircle_tp_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is None:
        _pt = point.Point(_x, _y)
    tool.pushObject(_pt)
    if len(tool) == 1:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", ccircle_tp_motion_notify_cb)
        gtkimage.setPrompt("Enter the second point or click in the drawing area")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)
    else:
        ccircle_tp_create(gtkimage, tool)

def ccircle_tpmode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", ccircle_tp_button_press_cb)
    tool.setHandler("entry_event", ccircle_tp_entry_make_pt)
    tool.setHandler("initialize", ccircle_tpmode_init)

#
# perpendicular construction line creation
#

def perp_cline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _hits = _active_layer.mapPoint((_x, _y), _tol, 1)
    if len(_hits):
        _obj, _pt = _hits[0]
        _lp = _active_layer.findObject(_pt)
        if _lp is None:
            _lp = _pt
            _active_layer.addObject(_lp)
        _redraw = True
        if isinstance(_obj, segment.Segment):
            _p1, _p2 = _obj.getEndpoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            if abs(_p1x - _p2x) < 1e-10: # vertical
                _active_layer.addObject(hcline.HCLine(_lp))
            elif abs(_p1y - _p2y) < 1e-10: # horizontal
                _active_layer.addObject(vcline.VCLine(_lp))
            else:
                _slope = (180.0/math.pi) * math.atan2((_p2y - _p1y),
                                                      (_p2x - _p1x)) + 90.0
                _active_layer.addObject(acline.ACLine(_lp, _slope))
        elif isinstance(_obj, circle._Circle):
            _cp = _obj.getCenter()
            _active_layer.addObject(cline.CLine(_cp, _lp))
        elif isinstance(_obj, hcline.HCLine):
            _active_layer.addObject(vcline.VCLine(_lp))
        elif isinstance(_obj, vcline.VCLine):
            _active_layer.addObject(hcline.HCLine(_lp))
        elif isinstance(_obj, acline.ACLine):
            _angle = _obj.getAngle()
            if abs(_angle) < 1e-10: # horizontal
                _active_layer.addObject(vcline.VCLine(_lp))
            elif abs(abs(_angle) - 90.0) < 1e-10: # vertical
                _active_layer.addObject(hcline.HCLine(_lp))
            else:
                _slope = _angle + 90.0
                _active_layer.addObject(acline.ACLine(_lp, _slope))
        elif isinstance(_obj, cline.CLine):
            _p1, _p2 = _obj.getKeypoints()
            _p1x, _p1y = _p1.getCoords()
            _p2x, _p2y = _p2.getCoords()
            if abs(_p1x - _p2x) < 1e-10: # vertical
                _active_layer.addObject(hcline.HCLine(_lp))
            elif abs(_p1y - _p2y) < 1e-10: # horizontal
                _active_layer.addObject(vcline.VCLine(_lp))
            else:
                _slope = (180.0/math.pi) * math.atan2((_p2y - _p1y),
                                                      (_p2x - _p1x)) + 90.0
                _active_layer.addObject(acline.ACLine(_lp, _slope))
        else:
            _active_layer.delObject(_lp)
            _redraw = False
        if _redraw:
            gtkimage.redraw()

def perpendicular_cline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", perp_cline_button_press_cb)

#
# tangent cline creation
#

def tangent_cline_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt, _new_pt = gtkimage.findPoint(_x, _y, _tol)
    if _pt is not None:
        _x, _y = _pt.getCoords()
        _active_layer = gtkimage.getActiveLayer()
        _layers = []
        _layers.append(gtkimage.getTopLayer())
        _rtd = 180.0/math.pi
        while len(_layers):
            _layer = _layers.pop()
            _cobj = None
            _angle = None
            for _circ in _layer.getLayerEntities("circle"):
                _cx, _cy = _circ.getCenter().getCoords()
                _rad = _circ.getRadius()
                if (_cx - _rad) > _x:
                    break
                if math.hypot((_x - _cx), (_y - _cy)) - _rad < 1e-10:
                    _cobj = _circ
                    _angle = _rtd * math.atan2((_y - _cy), (_x - _cx))
                    if _angle < 0.0:
                        _angle = _angle + 360.0
                    break
            if _cobj is None:
                for _arc in _layer.getLayerEntities("arc"):
                    _cx, _cy = _arc.getCenter().getCoords()
                    _rad = _arc.getRadius()
                    if (_cx - _rad) > _x:
                        break
                    if math.hypot((_cx - _x), (_cy - _y)) - _rad < 1e-10:
                        _angle = _rtd * math.atan2((_y - _cy), (_x - _cx))
                        if _angle < 0.0:
                            _angle = _angle + 360.0
                        if _arc.throughAngle(_angle):
                            _cobj = _arc
                            break
            if _cobj is not None:
                if _new_pt:
                    _active_layer.addObject(_pt)
                if (abs(_angle) < 1e-6 or
                    abs(_angle - 180.0) < 1e-6 or
                    abs(_angle - 360.0) < 1e-6):
                    _tcl = vcline.VCLine(_pt)
                elif (abs(_angle - 90.0) < 1e-6 or
                      abs(_angle - 270.0) < 1e-6):
                    _tcl = hcline.HCLine(_pt)
                else:
                    _slope = _angle + 90.0
                    _tcl = acline.ACLine(_pt, _slope)
                _active_layer.addObject(_tcl)
                gtkimage.redraw()
                break

def tangent_cline_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", tangent_cline_button_press_cb)
