#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# gtk dimension code
#
import math
import array

import gtk
import pango

import Generic.point as point
import Generic.arc as arc
import Generic.dimension as dimension


def set_ldim_prefix_suffix(ldim, gtkimage):
    assert isinstance(ldim, dimension.LinearDimension), "Not LinearDimension!"
    _pds = ldim.getPrimaryDimstring()
    _pds.setPrefix(gtkimage.getOption('DIM_PRIMARY_PREFIX'))
    _pds.setSuffix(gtkimage.getOption('DIM_PRIMARY_SUFFIX'))
    _sds = ldim.getSecondaryDimstring()
    _sds.setPrefix(gtkimage.getOption('DIM_SECONDARY_PREFIX'))
    _sds.setSuffix(gtkimage.getOption('DIM_SECONDARY_SUFFIX'))

def set_rdim_prefix_suffix(rdim, gtkimage):
    assert isinstance(rdim, dimension.RadialDimension), "Not RadialDimension!"
    _pds = rdim.getPrimaryDimstring()
    _pds.setPrefix(gtkimage.getOption('RADIAL_DIM_PRIMARY_PREFIX'))
    _pds.setSuffix(gtkimage.getOption('RADIAL_DIM_PRIMARY_SUFFIX'))
    _sds = rdim.getSecondaryDimstring()
    _sds.setPrefix(gtkimage.getOption('RADIAL_DIM_SECONDARY_PREFIX'))
    _sds.setSuffix(gtkimage.getOption('RADIAL_DIM_SECONDARY_SUFFIX'))

def set_adim_prefix_suffix(adim, gtkimage):
    assert isinstance(adim, dimension.AngularDimension), "Not AngularDimension!"
    _pds = adim.getPrimaryDimstring()
    _pds.setPrefix(gtkimage.getOption('ANGULAR_DIM_PRIMARY_PREFIX'))
    _pds.setSuffix(gtkimage.getOption('ANGULAR_DIM_PRIMARY_SUFFIX'))
    _sds = adim.getSecondaryDimstring()
    _sds.setPrefix(gtkimage.getOption('ANGULAR_DIM_SECONDARY_PREFIX'))
    _sds.setSuffix(gtkimage.getOption('ANGULAR_DIM_SECONDARY_SUFFIX'))
    
def set_dim_prefs(dim, gtkimage):
    _pds = dim.getPrimaryDimstring()
    _pds.setFamily(gtkimage.getOption('DIM_PRIMARY_FONT_FAMILY'))
    _pds.setSize(gtkimage.getOption('DIM_PRIMARY_FONT_SIZE'))
    _pds.setWeight(gtkimage.getOption('DIM_PRIMARY_FONT_WEIGHT'))
    _pds.setStyle(gtkimage.getOption('DIM_PRIMARY_FONT_STYLE'))
    _pds.setColor(gtkimage.getOption('DIM_PRIMARY_FONT_COLOR'))
    _pds.setPrecision(gtkimage.getOption('DIM_PRIMARY_PRECISION'))
    _pds.setUnits(gtkimage.getOption('DIM_PRIMARY_UNITS'))
    _pds.setPrintZero(gtkimage.getOption('DIM_PRIMARY_LEADING_ZERO'))
    _pds.setPrintDecimal(gtkimage.getOption('DIM_PRIMARY_TRAILING_DECIMAL'))
    _sds = dim.getSecondaryDimstring()
    _sds.setFamily(gtkimage.getOption('DIM_SECONDARY_FONT_FAMILY'))
    _sds.setSize(gtkimage.getOption('DIM_SECONDARY_FONT_SIZE'))
    _sds.setWeight(gtkimage.getOption('DIM_SECONDARY_FONT_WEIGHT'))
    _sds.setStyle(gtkimage.getOption('DIM_SECONDARY_FONT_STYLE'))
    _sds.setColor(gtkimage.getOption('DIM_SECONDARY_FONT_COLOR'))
    _sds.setPrecision(gtkimage.getOption('DIM_SECONDARY_PRECISION'))
    _sds.setUnits(gtkimage.getOption('DIM_SECONDARY_UNITS'))
    _sds.setPrintZero(gtkimage.getOption('DIM_SECONDARY_LEADING_ZERO'))
    _sds.setPrintDecimal(gtkimage.getOption('DIM_SECONDARY_TRAILING_DECIMAL'))
    dim.setOffset(gtkimage.getOption('DIM_OFFSET'))
    dim.setExtension(gtkimage.getOption('DIM_EXTENSION'))
    dim.setColor(gtkimage.getOption('DIM_COLOR'))
    dim.setPosition(gtkimage.getOption('DIM_POSITION'))
    dim.setEndpointType(gtkimage.getOption('DIM_ENDPOINT'))
    dim.setEndpointSize(gtkimage.getOption('DIM_ENDPOINT_SIZE'))
    dim.setDualDimMode(gtkimage.getOption('DIM_DUAL_MODE'))

#
# format the dimesion text string
#

def format_dim_layout(layout, ds):
    _attlist = pango.AttrList()
    _family = pango.AttrFamily(ds.getFamily())
    _family.start_index = 0
    _family.end_index = -1
    _attlist.insert(_family)
    _sty = pango.AttrStyle(ds.getStyle())
    _sty.start_index = 0
    _sty.end_index = -1
    _attlist.insert(_sty)
    _wt = pango.AttrWeight(ds.getWeight())
    _wt.start_index = 0
    _wt.end_index = -1
    _attlist.insert(_wt)
    _r, _g, _b = ds.getColor().getColors()
    _rs = int(65535 * (_r/255.0))
    _gs = int(65535 * (_g/255.0))
    _bs = int(65535 * (_b/255.0))
    _fg = pango.AttrForeground(_rs, _gs, _bs)
    _fg.start_index = 0
    _fg.end_index = -1
    _attlist.insert(_fg)
    _sz = pango.AttrSize(1024 * ds.getSize()) # need to scale
    _sz.start_index = 0
    _sz.end_index = -1
    _attlist.insert(_sz)
    layout.set_attributes(_attlist)

#
# draw dimension endpoints
#

def _calc_ldim_endpoint_pts(dim):
    _type = dim.getEndpointType()
    _dimclass = dimension.Dimension
    _epts = array.array('d')
    if _type != _dimclass.NO_ENDPOINT and _type != _dimclass.CIRCLE:
        _size = dim.getEndpointSize()
        _mpts = dim.getDimMarkerPoints()
        _x1 = _mpts[0]
        _y1 = _mpts[1]
        _x2 = _mpts[2]
        _y2 = _mpts[3]
        # print "x1: %g" % _x1
        # print "y1: %g" % _y1
        # print "x2: %g" % _x2
        # print "y2: %g" % _y2
        if abs(_x2 - _x1) < 1e-10: # vertical
            if _y2 > _y1:
                _sine = 1.0
                _cosine = 0.0
            else:
                _sine = -1.0
                _cosine = 0.0
        elif abs(_y2 - _y1) < 1e-10: # horizontal
            if _x2 > _x1:
                _sine = 0.0
                _cosine = 1.0
            else:
                _sine = 0.0
                _cosine = -1.0
        else:
            _angle = math.atan2((_y2 - _y1), (_x2 - _x1))
            _sine = math.sin(_angle)
            _cosine = math.cos(_angle)
        if _type == _dimclass.ARROW or _type == _dimclass.FILLED_ARROW:
            _height = _size/5.0
            # p1 -> (x,y) = (-size, _height)
            _ax1 = (_cosine * (-_size) - _sine * _height) + _x2
            _ay1 = (_sine * (-_size) + _cosine * _height) + _y2
            # p2 -> (x,y) = (-size, -_height)
            _ax2 = (_cosine * (-_size) - _sine *(-_height)) + _x2
            _ay2 = (_sine * (-_size) + _cosine *(-_height)) + _y2
            # p3 -> (x,y) = (size, _height)
            _ax3 = (_cosine * _size - _sine * _height) + _x1
            _ay3 = (_sine * _size + _cosine * _height) + _y1
            # p4 -> (x,y) = (size, -_height)
            _ax4 = (_cosine * _size - _sine *(-_height)) + _x1
            _ay4 = (_sine * _size + _cosine *(-_height)) + _y1
        elif _type == _dimclass.SLASH:
            _angle = 30.0 * (math.pi/180.0) # slope of slash
            _height = 0.5 * _size * math.sin(_angle)
            _length = 0.5 * _size * math.cos(_angle)
            # p1 -> (x,y) = (-_length, -_height)
            _sx1 = (_cosine * (-_length) - _sine * (-_height))
            _sy1 = (_sine * (-_length) + _cosine * (-_height))
            # p2 -> (x,y) = (_length, _height)
            _sx2 = (_cosine * _length - _sine * _height)
            _sy2 = (_sine * _length + _cosine * _height)
            #
            # shift the calculate based on the location of the
            # marker point
            #
            _ax1 = _sx1 + _x2
            _ay1 = _sy1 + _y2
            _ax2 = _sx2 + _x2
            _ay2 = _sy2 + _y2
            _ax3 = _sx1 + _x1
            _ay3 = _sy1 + _y1
            _ax4 = _sx2 + _x1
            _ay4 = _sy2 + _y1
        else:
            raise ValueError, "Unexpected endpoint type: '%s'" % str(_type)
        _epts.append(_ax1)
        _epts.append(_ay1)
        _epts.append(_ax2)
        _epts.append(_ay2)
        _epts.append(_ax3)
        _epts.append(_ay3)
        _epts.append(_ax4)
        _epts.append(_ay4)
        # for _ept in _epts:
            # print "ept: %g" % _ept
    return _epts

def _calc_rdim_endpoint_pts(rdim):
    _type = rdim.getEndpointType()
    _dimclass = dimension.Dimension
    _epts = array.array('d')
    if _type != _dimclass.NO_ENDPOINT and _type != _dimclass.CIRCLE:
        _size = rdim.getEndpointSize()
        _mpts = rdim.getDimMarkerPoints()
        _dia_mode = rdim.getDiaMode()
        _x1 = _mpts[0]
        _y1 = _mpts[1]
        _x2, _y2 = rdim.getDimCircle().getCenter().getCoords()        
        if _dia_mode:
            _x2 = _mpts[2]
            _y2 = _mpts[3]
        print "x1: %g" % _x1
        print "y1: %g" % _y1
        print "x2: %g" % _x2
        print "y2: %g" % _y2
        if abs(_x2 - _x1) < 1e-10: # vertical
            if _y2 > _y1:
                _sine = 1.0
                _cosine = 0.0
            else:
                _sine = -1.0
                _cosine = 0.0
        elif abs(_y2 - _y1) < 1e-10: # horizontal
            if _x2 > _x1:
                _sine = 0.0
                _cosine = 1.0
            else:
                _sine = 0.0
                _cosine = -1.0
        else:
            _angle = math.atan2((_y2 - _y1), (_x2 - _x1))
            _sine = math.sin(_angle)
            _cosine = math.cos(_angle)
        if _type == _dimclass.ARROW or _type == _dimclass.FILLED_ARROW:
            _height = _size/5.0
            # p1 -> (x,y) = (size, _height)
            _ax1 = (_cosine * _size - _sine * _height) + _x1
            _ay1 = (_sine * _size + _cosine * _height) + _y1
            # p2 -> (x,y) = (size, -_height)
            _ax2 = (_cosine * _size - _sine *(-_height)) + _x1
            _ay2 = (_sine * _size + _cosine *(-_height)) + _y1
            if _dia_mode:
                # p3 -> (x,y) = (-size, _height)
                _ax3 = (_cosine * (-_size) - _sine * _height) + _x2
                _ay3 = (_sine * (-_size) + _cosine * _height) + _y2
                # p4 -> (x,y) = (-size, -_height)
                _ax4 = (_cosine * (-_size) - _sine *(-_height)) + _x2
                _ay4 = (_sine * (-_size) + _cosine *(-_height)) + _y2
        elif _type == _dimclass.SLASH:
            _angle = 30.0 * (math.pi/180.0) # slope of slash
            _height = 0.5 * _size * math.sin(_angle)
            _length = 0.5 * _size * math.cos(_angle)
            # p1 -> (x,y) = (-_length, -_height)
            _sx1 = (_cosine * (-_length) - _sine * (-_height))
            _sy1 = (_sine * (-_length) + _cosine * (-_height))
            # p2 -> (x,y) = (_length, _height)
            _sx2 = (_cosine * _length - _sine * _height)
            _sy2 = (_sine * _length + _cosine * _height)
            #
            # shift the calculate based on the location of the
            # marker point
            #
            _ax1 = _sx1 + _x1
            _ay1 = _sy1 + _y1
            _ax2 = _sx2 + _x1
            _ay2 = _sy2 + _y1
            if _dia_mode:
                _ax3 = _sx1 + _x2
                _ay3 = _sy1 + _y2
                _ax4 = _sx2 + _x2
                _ay4 = _sy2 + _y2
        else:
            raise ValueError, "Unexpected endpoint type: '%s'" % str(_type)
        _epts.append(_ax1)
        _epts.append(_ay1)
        _epts.append(_ax2)
        _epts.append(_ay2)
        if _dia_mode:
            _epts.append(_ax3)
            _epts.append(_ay3)
            _epts.append(_ax4)
            _epts.append(_ay4)
        # for _ept in _epts:
            # print "ept: %g" % _ept
    return _epts

def _calc_adim_endpoint_pts(adim):
    _type = adim.getEndpointType()
    _dimclass = dimension.Dimension
    _epts = array.array('d')
    if _type != _dimclass.NO_ENDPOINT and _type != _dimclass.CIRCLE:
        _size = adim.getEndpointSize()
        _da1, _da2 = adim.getDimAngles()
        # print "da1: %g" % _da1
        # print "da2: %g" % _da2
        _a1 = _da1 - 90.0
        _a2 = _da2 - 90.0
        # print "a1: %g" % _a1
        # print "a2: %g" % _a2
        _mpts = adim.getDimMarkerPoints()
        _x1 = _mpts[0]
        _y1 = _mpts[1]
        _x2 = _mpts[2]
        _y2 = _mpts[3]
        print "x1: %g" % _x1
        print "y1: %g" % _y1
        print "x2: %g" % _x2
        print "y2: %g" % _y2
        _angle = math.atan2((_y2 - _y1), (_x2 - _x1))
        _dtr = math.pi/180.0
        _sin1 = math.sin(_dtr * _a1)
        _cos1 = math.cos(_dtr * _a1)
        _sin2 = math.sin(_dtr * _a2)
        _cos2 = math.cos(_dtr * _a2)
        if _type == _dimclass.ARROW or _type == _dimclass.FILLED_ARROW:
            _height = _size/5.0
            # p1 -> (x,y) = (size, _height)
            _ax1 = (_cos1 * (-_size) - _sin1 * _height) + _x1
            _ay1 = (_sin1 * (-_size) + _cos1 * _height) + _y1
            # p2 -> (x,y) = (size, -_height)
            _ax2 = (_cos1 * (-_size) - _sin1 *(-_height)) + _x1
            _ay2 = (_sin1 * (-_size) + _cos1 *(-_height)) + _y1
            # p3 -> (x,y) = (size, _height)
            _ax3 = (_cos2 * _size - _sin2 * _height) + _x2
            _ay3 = (_sin2 * _size + _cos2 * _height) + _y2
            # p4 -> (x,y) = (size, -_height)
            _ax4 = (_cos2 * _size - _sin2 *(-_height)) + _x2
            _ay4 = (_sin2 * _size + _cos2 *(-_height)) + _y2
        elif _type == _dimclass.SLASH:
            _angle = 30.0 * (math.pi/180.0) # slope of slash
            _height = 0.5 * _size * math.sin(_angle)
            _length = 0.5 * _size * math.cos(_angle)
            # p1 -> (x,y) = (-_length, -_height)
            _ax1 = (_cos1 * (-_length) - _sin1 * (-_height)) + _x1
            _ay1 = (_sin1 * (-_length) + _cos1 * (-_height)) + _y1
            # p2 -> (x,y) = (_length, _height)
            _ax2 = (_cos1 * _length - _sin1 * _height) + _x1
            _ay2 = (_sin1 * _length + _cos1 * _height) + _y1
            # p3 -> (x,y) = (-_length, -_height)
            _ax3 = (_cos2 * (-_length) - _sin2 * (-_height)) + _x2
            _ay3 = (_sin2 * (-_length) + _cos2 * (-_height)) + _y2
            # p4 -> (x,y) = (_length, _height)
            _ax4 = (_cos2 * _length - _sin2 * _height) + _x2
            _ay4 = (_sin2 * _length + _cos2 * _height) + _y2            
        else:
            raise ValueError, "Unexpected endpoint type: '%s'" % str(_type)
        _epts.append(_ax1)
        _epts.append(_ay1)
        _epts.append(_ax2)
        _epts.append(_ay2)
        _epts.append(_ax3)
        _epts.append(_ay3)
        _epts.append(_ax4)
        _epts.append(_ay4)
        # for _ept in _epts:
            # print "ept: %g" % _ept
    return _epts

#
# linear dimension motion-notify handler
#

def dim_pts_motion_notify_cb(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    _gc = gtkimage.getGC()
    _x = event.x
    _y = event.y
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        widget.window.draw_line(_gc, _px, _py, _xc, _yc)
    tool.setCurrentPoint(_x, _y)
    widget.window.draw_line(_gc, _px, _py, _x, _y)

def dim_txt_motion_notify_cb(gtkimage, widget, event, tool):
    assert len(tool) == 1, "Unexpected tool count"
    _ix, _iy = gtkimage.getPoint()
    _gc = gtkimage.getGC()
    _ex = event.x
    _ey = event.y
    _cp = tool.getCurrentPoint()
    _dim = tool.getObject(0)
    if _cp is None:
        _fgc = widget.get_style().fg_gc[gtk.STATE_NORMAL]
        _x, _y = tool.getCoordPair(0)
        _px1, _py1 = gtkimage.coordToPixTransform(_x, _y)
        _x, _y = tool.getCoordPair(1)
        _px2, _py2 = gtkimage.coordToPixTransform(_x, _y)
        widget.window.draw_line(_fgc, _px1, _py1, _px2, _py2)
        # _black_gc = widget.get_style().black_gc
        # _win = widget.window
        # _width, _height = _win.get_size()
        # _win.draw_rectangle(_black_gc, gtk.TRUE, 0, 0, _width, _height)
        # widget.queue_draw()
    else:
        _dxpts = _dim.getDimXPoints()
        _dypts = _dim.getDimYPoints()
        _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
        _p1x, _p1y = gtkimage.coordToPixTransform(_dxpts[1], _dypts[1])
        _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[2], _dypts[2])
        _p3x, _p3y = gtkimage.coordToPixTransform(_dxpts[3], _dypts[3])
        _p4x, _p4y = gtkimage.coordToPixTransform(_dxpts[4], _dypts[4])
        _p5x, _p5y = gtkimage.coordToPixTransform(_dxpts[5], _dypts[5])
        widget.window.draw_line(_gc, _p0x, _p0y, _p2x, _p2y)
        widget.window.draw_line(_gc, _p3x, _p3y, _p5x, _p5y)
        widget.window.draw_line(_gc, _p1x, _p1y, _p4x, _p4y)
    tool.setCurrentPoint(_ex, _ey)
    _dim.setLocation(_ix, _iy)
    _dxpts = _dim.getDimXPoints()
    _dypts = _dim.getDimYPoints()
    _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
    _p1x, _p1y = gtkimage.coordToPixTransform(_dxpts[1], _dypts[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[2], _dypts[2])
    _p3x, _p3y = gtkimage.coordToPixTransform(_dxpts[3], _dypts[3])
    _p4x, _p4y = gtkimage.coordToPixTransform(_dxpts[4], _dypts[4])
    _p5x, _p5y = gtkimage.coordToPixTransform(_dxpts[5], _dypts[5])
    widget.window.draw_line(_gc, _p0x, _p0y, _p2x, _p2y)
    widget.window.draw_line(_gc, _p3x, _p3y, _p5x, _p5y)
    widget.window.draw_line(_gc, _p1x, _p1y, _p4x, _p4y)

#
# draw linear, horizontal, and vertical dimensions
#

def draw_linear_dimension(gtkimage, dim, active):
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                            gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.COPY)
    _da = gtkimage.getDA()
    _window = _da.window
    _ctx = _da.get_pango_context()
    _width, _height = gtkimage.getSize()
    _pixmap = gtkimage.getPixmap()
    _dxpts = dim.getDimXPoints()
    _dypts = dim.getDimYPoints()
    _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
    _p1x, _p1y = gtkimage.coordToPixTransform(_dxpts[1], _dypts[1])
    _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[2], _dypts[2])
    _p3x, _p3y = gtkimage.coordToPixTransform(_dxpts[3], _dypts[3])
    _p4x, _p4y = gtkimage.coordToPixTransform(_dxpts[4], _dypts[4])
    _p5x, _p5y = gtkimage.coordToPixTransform(_dxpts[5], _dypts[5])
    if active:
        _dim_color = dim.getColor()
    else:
        _dim_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    if gtkimage.hasGTKColor(_dim_color):
        _color = gtkimage.getGTKColor(_dim_color)
    else:
        _cstring = str(_dim_color)
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_dim_color, _color)
    _gc.set_foreground(_color)
    #
    # draw dimension lines
    #
    _pixmap.draw_line(_gc, _p0x, _p0y, _p2x, _p2y)
    _pixmap.draw_line(_gc, _p3x, _p3y, _p5x, _p5y)
    _pixmap.draw_line(_gc, _p1x, _p1y, _p4x, _p4y)
    #
    # draw endpoints
    #
    _etype = dim.getEndpointType()
    _mpts = dim.getDimMarkerPoints()
    _mp1x, _mp1y = gtkimage.coordToPixTransform(_mpts[0], _mpts[1])
    _mp2x, _mp2y = gtkimage.coordToPixTransform(_mpts[2], _mpts[3])
    if (_etype == dimension.Dimension.ARROW or
        _etype == dimension.Dimension.FILLED_ARROW or
        _etype == dimension.Dimension.SLASH):
        _epts = _calc_ldim_endpoint_pts(dim)
        assert len(_epts) == 8, "Unexpected endpoint array length: %d" % len(_epts)        
        if _etype == dimension.Dimension.ARROW:
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _pixmap.draw_line(_gc, _epx1, _epy1, _mp2x, _mp2y)
            _pixmap.draw_line(_gc, _epx2, _epy2, _mp2x, _mp2y)
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
            _pixmap.draw_line(_gc, _epx1, _epy1, _mp1x, _mp1y)
            _pixmap.draw_line(_gc, _epx2, _epy2, _mp1x, _mp1y)            
        elif _etype == dimension.Dimension.FILLED_ARROW:
            _tuples = []
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _tuples.append((_epx1, _epy1))
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _tuples.append((_epx2, _epy2))
            _tuples.append((_mp2x, _mp2y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
            del _tuples[:]
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
            _tuples.append((_epx1, _epy1))
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
            _tuples.append((_epx2, _epy2))
            _tuples.append((_mp1x, _mp1y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
        elif _etype == dimension.Dimension.SLASH:
            print "add slash endpoint drawing ..."
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _pixmap.draw_line(_gc, _epx1, _epy1, _epx2, _epy2)
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
            _pixmap.draw_line(_gc, _epx1, _epy1, _epx2, _epy2)
    elif _etype == dimension.Dimension.CIRCLE:
        _upp = gtkimage.getUnitsPerPixel()
        _size = dim.getEndpointSize()
        _width = int(_size/_upp)
        _cw = _ch = _width * 2
        _xmin = _mp1x - _width
        _ymin = _mp1y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
        _xmin = _mp2x - _width
        _ymin = _mp2y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
    else:
        print "no endpoint on dimension ..."
    #
    # draw dimension text
    #
    _dx, _dy = dim.getLocation()
    _pdx, _pdy = gtkimage.coordToPixTransform(_dx, _dy)    
    _dlen = dim.calculate()
    _hlen = gtkimage.scaleLength(_dlen)
    _dims = dim.getDimensions(_hlen)
    _dlines = []
    _pri_layout = pango.Layout(_ctx)
    try:
        _pri_layout.set_text(_dims[0])
    except:
        _pri_layout.set_text(_dims[0], len(_dims[0]))
    _ds = dim.getPrimaryDimstring()
    format_dim_layout(_pri_layout, _ds)
    _pri_w, _pri_h = _pri_layout.get_pixel_size()
    _layout_width = _pri_w + 2
    _layout_height = _pri_h + 2
    _dual_mode = dim.getDualDimMode()
    if _dual_mode:
        _sec_layout = pango.Layout(_ctx)
        try:
            _sec_layout.set_text(_dims[1])
        except:
            _sec_layout.set_text(_dims[1], len(_dims[1]))
        _sds = dim.getSecondaryDimstring()
        format_dim_layout(_sec_layout, _sds)
        _sec_w, _sec_h = _sec_layout.get_pixel_size()
        _layout_width = max(_pri_w, _sec_w) + 2 # extra for padding space
        _layout_height = _pri_h + _sec_h + 2 # same
    _pos = dim.getPosition() # fixme
    _hw = _layout_width//2
    _hh = _layout_height//2
    #
    # fill over the area where the text will be drawn
    #
    _box_gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _pixmap.draw_rectangle(_box_gc, gtk.TRUE, (_pdx - _hw), (_pdy - _hh),
                           _layout_width, _layout_height)
    del _box_gc # needed???
    #
    # draw the dimension text
    #
    _pgc = _da.get_style().black_gc
    if _dual_mode:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - _pri_h - 1), _pri_layout)
        _pixmap.draw_layout(_pgc, (_pdx - (_sec_w//2)), (_pdy + 1), _sec_layout)
        _pixmap.draw_line(_gc, (_pdx - _hw + 1), _pdy, (_pdx + _hw - 1), _pdy)
    else:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - (_pri_h//2)), _pri_layout)

def add_dimension(gtkimage, widget, event, tool):
    assert len(tool) == 1, "Unexpected tool count"
    _dim = tool.popObject()
    _x, _y = gtkimage.getPoint()
    _dim.setLocation(_x, _y)
    set_dim_prefs(_dim, gtkimage)
    if isinstance(_dim, dimension.LinearDimension):
        set_ldim_prefix_suffix(_dim, gtkimage)
    elif isinstance(_dim, dimension.RadialDimension):
        set_rdim_prefix_suffix(_dim, gtkimage)
    elif isinstance(_dim, dimension.AngularDimension):
        set_adim_prefix_suffix(_dim, gtkimage)
    try:
        gtkimage.addObject(_dim)
    except StandardError, e:
        raise StandardError, "Failed to add Dimension:" + str(e)
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkimage.redraw()
    init_handler = tool.getHandler("initialize")
    init_handler(tool)

def linear_pts_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _pt = None
    _layer = None
    _layers = []
    _layers.append(gtkimage.getTopLayer())
    while(len(_layers)):
        _layer = _layers.pop()
        _pt = _layer.hasPoint((_x, _y), _tol)
        if _pt is not None:
            break
        if _layer.getVisibility():
            for _child in _layer.getChildren():
                _layers.append(_child)
    if _pt is not None:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        if not len(tool): # first endpoint
            tool.pushObject(_layer)
            tool.pushObject(_pt)
            tool.setLocation(_x, _y)
            tool.setHandler("motion_notify", dim_pts_motion_notify_cb)
            gtkimage.setPrompt("Click on the second point for the dimension.")
        else: # second endpoint
            _p1 = tool.popObject()
            _l1 = tool.popObject()
            _ds = gtkimage.getOption('DIM_STYLE')
            _dimtype = tool.getObjType()
            _dim = _dimtype(_l1, _p1, _layer, _pt, _x, _y, _ds)
            tool.pushObject(_dim)
            tool.clearCurrentPoint()            
            tool.setHandler("motion_notify", dim_txt_motion_notify_cb)
            tool.setHandler("button_press", add_dimension)
            gtkimage.refresh()
            gtkimage.setPrompt("Click where the dimension text should go.")
        gtkimage.getGC().set_function(gtk.gdk.INVERT)            

#
# linear dimensions
#

def linear_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", linear_pts_button_press_cb)
    tool.setObjType(dimension.LinearDimension)
    tool.setHandler("initialize", linear_mode_init)

#
# horizontal dimensions
#


def horizontal_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", linear_pts_button_press_cb)
    tool.setObjType(dimension.HorizontalDimension)
    tool.setHandler("initialize", horizontal_mode_init)

#
# vertical dimensions
#


def vertical_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", linear_pts_button_press_cb)
    tool.setObjType(dimension.VerticalDimension)
    tool.setHandler("initialize", vertical_mode_init)

#
# radial dimensions
#

def rdim_txt_motion_notify_cb(gtkimage, widget, event, tool):
    assert len(tool) == 1, "Unexpected tool count"
    _x, _y = gtkimage.getPoint()
    _gc = gtkimage.getGC()
    _ex = event.x
    _ey = event.y
    _cp = tool.getCurrentPoint()
    _rdim = tool.getObject(0)
    _win = widget.window
    if _cp is not None:
        _dxpts = _rdim.getDimXPoints()
        _dypts = _rdim.getDimYPoints()
        _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
        _p1x, _p1y = gtkimage.coordToPixTransform(_dxpts[1], _dypts[1])
        _win.draw_line(_gc, _p0x, _p0y, _p1x, _p1y)
    tool.setCurrentPoint(_ex, _ey)
    _rdim.setLocation(_x, _y)
    _dxpts = _rdim.getDimXPoints()
    _dypts = _rdim.getDimYPoints()
    _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
    _p1x, _p1y = gtkimage.coordToPixTransform(_dxpts[1], _dypts[1])
    _win.draw_line(_gc, _p0x, _p0y, _p1x, _p1y)

def draw_radial_dimension(gtkimage, rdim, active):
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                            gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.COPY)
    _da = gtkimage.getDA()
    _window = _da.window
    _ctx = _da.get_pango_context()
    _width, _height = gtkimage.getSize()
    _pixmap = gtkimage.getPixmap()
    _dxpts = rdim.getDimXPoints()
    _dypts = rdim.getDimYPoints()
    _p1x, _p1y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
    _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[1], _dypts[1])
    _dx, _dy = rdim.getLocation()
    _pdx, _pdy = gtkimage.coordToPixTransform(_dx, _dy)
    if active:
        _dim_color = rdim.getColor()
    else:
        _dim_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    if gtkimage.hasGTKColor(_dim_color):
        _color = gtkimage.getGTKColor(_dim_color)
    else:
        _cstring = str(_dim_color)
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_dim_color, _color)
    _gc.set_foreground(_color)
    _pixmap.draw_line(_gc, _p1x, _p1y, _p2x, _p2y)
    #
    # draw marker points
    #
    _mpts = rdim.getDimMarkerPoints()
    _mp1x, _mp1y = gtkimage.coordToPixTransform(_mpts[0], _mpts[1])
    _etype = rdim.getEndpointType()
    _dia_mode = rdim.getDiaMode()
    if (_etype == dimension.Dimension.ARROW or
        _etype == dimension.Dimension.FILLED_ARROW or
        _etype == dimension.Dimension.SLASH):
        _epts = _calc_rdim_endpoint_pts(rdim)
        if _dia_mode:
            assert len(_epts) == 8, "Unexpected endpoint array length: %d" % len(_epts)
        else:
            assert len(_epts) == 4, "Unexpected endpoint array length: %d" % len(_epts)
        if _etype == dimension.Dimension.ARROW:
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _pixmap.draw_line(_gc, _epx1, _epy1, _mp1x, _mp1y)
            _pixmap.draw_line(_gc, _epx2, _epy2, _mp1x, _mp1y)
            if _dia_mode:
                _mp2x, _mp2y = gtkimage.coordToPixTransform(_mpts[2], _mpts[3])
                _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
                _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
                _pixmap.draw_line(_gc, _epx1, _epy1, _mp2x, _mp2y)
                _pixmap.draw_line(_gc, _epx2, _epy2, _mp2x, _mp2y)            
        elif _etype == dimension.Dimension.FILLED_ARROW:
            _tuples = []
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _tuples.append((_epx1, _epy1))
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _tuples.append((_epx2, _epy2))
            _tuples.append((_mp1x, _mp1y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
            if _dia_mode:
                _mp2x, _mp2y = gtkimage.coordToPixTransform(_mpts[2], _mpts[3])
                del _tuples[:]
                _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
                _tuples.append((_epx1, _epy1))
                _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
                _tuples.append((_epx2, _epy2))
                _tuples.append((_mp2x, _mp2y))
                _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
        elif _etype == dimension.Dimension.SLASH:
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _pixmap.draw_line(_gc, _epx1, _epy1, _epx2, _epy2)
            if _dia_mode:
                _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
                _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
                _pixmap.draw_line(_gc, _epx1, _epy1, _epx2, _epy2)
    elif _etype == dimension.Dimension.CIRCLE:
        _upp = gtkimage.getUnitsPerPixel()
        _size = rdim.getEndpointSize()
        _width = int(_size/_upp)
        _cw = _ch = _width * 2
        _xmin = _mp1x - _width
        _ymin = _mp1y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
        if _dia_mode:
            _mp2x, _mp2y = gtkimage.coordToPixTransform(_mpts[2], _mpts[3])
            _xmin = _mp2x - _width
            _ymin = _mp2y - _width
            _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                             0, (360 * 64))
    else:
        print "no endpoint on dimension ..."
    _rlen = gtkimage.scaleLength(rdim.calculate())
    _dims = rdim.getDimensions(_rlen)
    _pri_layout = pango.Layout(_ctx)
    try:
        _pri_layout.set_text(_dims[0])
    except:
        _pri_layout.set_text(_dims[0], len(_dims[0]))
    _pds = rdim.getPrimaryDimstring()
    format_dim_layout(_pri_layout, _pds)
    _pri_w, _pri_h = _pri_layout.get_pixel_size()
    _layout_width = _pri_w + 2
    _layout_height = _pri_h + 2
    _dual_mode = rdim.getDualDimMode()
    if _dual_mode:
        _sec_layout = pango.Layout(_ctx)
        try:
            _sec_layout.set_text(_dims[1])
        except:
            _sec_layout.set_text(_dims[1], len(_dims[1]))
        _sds = rdim.getSecondaryDimstring()
        format_dim_layout(_sec_layout, _sds)
        _sec_w, _sec_h = _sec_layout.get_pixel_size()
        _layout_width = max(_pri_w, _sec_w) + 2 # extra for padding space
        _layout_height = _pri_h + _sec_h + 2 # same
    _pos = rdim.getPosition() # fixme
    _hw = _layout_width//2
    _hh = _layout_height//2
    #
    # fill over the area where the text will be drawn
    #
    _box_gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _pixmap.draw_rectangle(_box_gc, gtk.TRUE, (_pdx - _hw), (_pdy - _hh),
                           _layout_width, _layout_height)
    del _box_gc # needed???
    #
    # draw the dimension text
    #
    _pgc = _da.get_style().black_gc
    if _dual_mode:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)),
                            (_pdy - _pri_h - 1), _pri_layout)
        _pixmap.draw_layout(_pgc, (_pdx - (_sec_w//2)),
                            (_pdy + 1), _sec_layout)
        _pixmap.draw_line(_gc, (_pdx - _hw), _pdy,
                          (_pdx + _hw), _pdy)
    else:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)),
                            (_pdy - (_pri_h//2)), _pri_layout)

def radial_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    if len(tool):
        _rdim = tool.popObject()
        _rdim.setLocation(_x, _y)
        set_dim_prefs(_rdim, gtkimage)
        set_rdim_prefix_suffix(_rdim, gtkimage)
        try:
            gtkimage.addObject(_rdim)
        except StandardError, e:
            raise StandardError, "Failed to add Dimension:" + str(e)
        gtkimage.setPrompt("Click on the circle or arc to be dimensioned.")
        gtkimage.redraw()
        radial_mode_init(tool)
    else:
        _dc = None
        _dl = None
        _layers = []
        _layers.append(gtkimage.getTopLayer())
        while(len(_layers)):
            _layer = _layers.pop()
            _cobjs = (_layer.getLayerEntities("circle") +
                      _layer.getLayerEntities("arc"))
            for _cobj in _cobjs:
                _mp = _cobj.mapPoint((_x, _y), _tol)
                if _mp is not None:
                    _dc = _cobj
                    _dl = _layer
                    break
            if _layer.getVisibility():
                for _child in _layer.getChildren():
                    _layers.append(_child)
        if _dc is not None:
            _ds = gtkimage.getOption('DIM_STYLE')
            _cx, _cy = _dc.getCenter().getCoords()
            tool.storeCoords(_cx, _cy)
            _rdim = dimension.RadialDimension(_dl, _dc, _x, _y, _ds)
            tool.pushObject(_rdim)
            tool.setHandler("motion_notify", rdim_txt_motion_notify_cb)
            gtkimage.setPrompt("Click where the dimension text should be placed.")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)

def radial_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", radial_button_press_cb)

#
# angular dimensions
#

def adim_txt_motion_notify_cb(gtkimage, widget, event, tool):
    assert len(tool) == 1, "Unexpected tool count"
    _ix, _iy = gtkimage.getPoint()
    _gc = gtkimage.getGC()
    _ex = event.x
    _ey = event.y
    _cp = tool.getCurrentPoint()
    _adim = tool.getObject(0)
    _vx, _vy = _adim.getVertexPoint().getCoords()
    _px, _py = gtkimage.coordToPixTransform(_vx, _vy)
    _win = widget.window
    if _cp is not None:
        _dxpts = _adim.getDimXPoints()
        _dypts = _adim.getDimYPoints()
        #
        # draw bars
        #
        _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
        _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[2], _dypts[2])
        _p3x, _p3y = gtkimage.coordToPixTransform(_dxpts[3], _dypts[3])
        _p5x, _p5y = gtkimage.coordToPixTransform(_dxpts[5], _dypts[5])
        _win.draw_line(_gc, _p0x, _p0y, _p2x, _p2y)
        _win.draw_line(_gc, _p3x, _p3y, _p5x, _p5y)
        #
        # draw arc
        #
        _angles = _adim.getDimAngles()
        _sa = _angles[0]
        _ea = _angles[1]
        _rad = int(_adim.getDimRadius()/gtkimage.getUnitsPerPixel())
        _pxmin = _px - _rad
        _pymin = _py - _rad
        _cw = _ch = _rad * 2
        if _sa < _ea:
            _sweep = _ea - _sa
            _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                          (_sa * 64), (_sweep * 64))
        else:
            _sweep = 360.0 - _sa
            _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                          (_sa * 64), (_sweep * 64))
            _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                          0, (_ea * 64))
    tool.setCurrentPoint(_ex, _ey)
    _adim.setLocation(_ix, _iy)
    _dxpts = _adim.getDimXPoints()
    _dypts = _adim.getDimYPoints()
    #
    # draw bars
    #
    _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
    _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[2], _dypts[2])
    _p3x, _p3y = gtkimage.coordToPixTransform(_dxpts[3], _dypts[3])
    _p5x, _p5y = gtkimage.coordToPixTransform(_dxpts[5], _dypts[5])
    _win.draw_line(_gc, _p0x, _p0y, _p2x, _p2y)
    _win.draw_line(_gc, _p3x, _p3y, _p5x, _p5y)
    #
    # draw arc
    #
    _angles = _adim.getDimAngles()
    _sa = _angles[0]
    _ea = _angles[1]
    _rad = int(_adim.getDimRadius()/gtkimage.getUnitsPerPixel())
    _pxmin = _px - _rad
    _pymin = _py - _rad
    _cw = _ch = _rad * 2
    if _sa < _ea:
        _sweep = _ea - _sa
        _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                      (_sa * 64), int(_sweep * 64))
    else:
        _sweep = 360.0 - _sa
        _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                      (_sa * 64), (_sweep * 64))
        _win.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                      0, (_ea * 64))
        
def draw_angular_dimension(gtkimage, adim, active):
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                            gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.COPY)
    _da = gtkimage.getDA()
    _window = _da.window
    _ctx = _da.get_pango_context()
    _width, _height = gtkimage.getSize()
    _pixmap = gtkimage.getPixmap()
    _dxpts = adim.getDimXPoints()
    _dypts = adim.getDimYPoints()
    _p0x, _p0y = gtkimage.coordToPixTransform(_dxpts[0], _dypts[0])
    _p2x, _p2y = gtkimage.coordToPixTransform(_dxpts[2], _dypts[2])
    _p3x, _p3y = gtkimage.coordToPixTransform(_dxpts[3], _dypts[3])
    _p5x, _p5y = gtkimage.coordToPixTransform(_dxpts[5], _dypts[5])
    if active:
        _dim_color = adim.getColor()
    else:
        _dim_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    if gtkimage.hasGTKColor(_dim_color):
        _color = gtkimage.getGTKColor(_dim_color)
    else:
        _cstring = str(_dim_color)
        _color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_dim_color, _color)
    _gc.set_foreground(_color)
    #
    # draw dimension lines
    #
    _pixmap.draw_line(_gc, _p0x, _p0y, _p2x, _p2y)
    _pixmap.draw_line(_gc, _p3x, _p3y, _p5x, _p5y)
    #
    # draw the dimension crossbar/arc
    #
    _vx, _vy = adim.getVertexPoint().getCoords()
    _px, _py = gtkimage.coordToPixTransform(_vx, _vy)
    _dx, _dy = adim.getLocation()
    _pdx, _pdy = gtkimage.coordToPixTransform(_dx, _dy)
    _angles = adim.getDimAngles()
    _rad = int(adim.getDimRadius()/gtkimage.getUnitsPerPixel())
    _pxmin = _px - _rad
    _pymin = _py - _rad
    _cw = _ch = _rad * 2
    _sa = _angles[0]
    _ea = _angles[1]
    if _sa < _ea:
        _sweep = _ea - _sa
        _pixmap.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                         (_sa * 64), (_sweep * 64))
    else:
        _sweep = 360.0 - _sa
        _pixmap.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                         (_sa * 64), (_sweep * 64))
        _pixmap.draw_arc(_gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                         0, (_ea * 64))
    #
    # draw endpont markers
    #
    _etype = adim.getEndpointType()
    _mpts = adim.getDimMarkerPoints()
    _mp1x, _mp1y = gtkimage.coordToPixTransform(_mpts[0], _mpts[1])
    _mp2x, _mp2y = gtkimage.coordToPixTransform(_mpts[2], _mpts[3])
    if (_etype == dimension.Dimension.ARROW or
        _etype == dimension.Dimension.FILLED_ARROW or
        _etype == dimension.Dimension.SLASH):
        _epts = _calc_adim_endpoint_pts(adim)
        assert len(_epts) == 8, "Unexpected endpoint array length: %d" % len(_epts)        
        if _etype == dimension.Dimension.ARROW:
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _pixmap.draw_line(_gc, _epx1, _epy1, _mp1x, _mp1y)
            _pixmap.draw_line(_gc, _epx2, _epy2, _mp1x, _mp1y)
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
            _pixmap.draw_line(_gc, _epx1, _epy1, _mp2x, _mp2y)
            _pixmap.draw_line(_gc, _epx2, _epy2, _mp2x, _mp2y)            
        elif _etype == dimension.Dimension.FILLED_ARROW:
            _tuples = []
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _tuples.append((_epx1, _epy1))
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _tuples.append((_epx2, _epy2))
            _tuples.append((_mp1x, _mp1y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
            del _tuples[:]
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
            _tuples.append((_epx1, _epy1))
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
            _tuples.append((_epx2, _epy2))
            _tuples.append((_mp2x, _mp2y))
            _pixmap.draw_polygon(_gc, gtk.TRUE, _tuples)
        elif _etype == dimension.Dimension.SLASH:
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[0], _epts[1])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[2], _epts[3])
            _pixmap.draw_line(_gc, _epx1, _epy1, _epx2, _epy2)
            _epx1, _epy1 =  gtkimage.coordToPixTransform(_epts[4], _epts[5])
            _epx2, _epy2 =  gtkimage.coordToPixTransform(_epts[6], _epts[7])
            _pixmap.draw_line(_gc, _epx1, _epy1, _epx2, _epy2)
    elif _etype == dimension.Dimension.CIRCLE:
        _upp = gtkimage.getUnitsPerPixel()
        _size = adim.getEndpointSize()
        _width = int(_size/_upp)
        _cw = _ch = _width * 2
        _xmin = _mp1x - _width
        _ymin = _mp1y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
        _xmin = _mp2x - _width
        _ymin = _mp2y - _width
        _pixmap.draw_arc(_gc, gtk.TRUE, _xmin, _ymin, _cw, _ch,
                         0, (360 * 64))
    else:
        print "no endpoint on dimension ..."
    #
    # draw dimension text
    #
    _dlen = adim.calculate()
    _dims = adim.getDimensions(_dlen)
    _dlines = []
    _pri_layout = pango.Layout(_ctx)
    try:
        _pri_layout.set_text(_dims[0])
    except:
        _pri_layout.set_text(_dims[0], len(_dims[0]))
    _ds = adim.getPrimaryDimstring()
    format_dim_layout(_pri_layout, _ds)
    _pri_w, _pri_h = _pri_layout.get_pixel_size()
    _layout_width = _pri_w + 2
    _layout_height = _pri_h + 2
    _dual_mode = adim.getDualDimMode()
    if _dual_mode:
        _sec_layout = pango.Layout(_ctx)
        try:
            _sec_layout.set_text(_dims[1])
        except:
            _sec_layout.set_text(_dims[1], len(_dims[1]))
        _sds = adim.getSecondaryDimstring()
        format_dim_layout(_sec_layout, _sds)
        _sec_w, _sec_h = _sec_layout.get_pixel_size()
        _layout_width = max(_pri_w, _sec_w) + 2 # extra for padding space
        _layout_height = _pri_h + _sec_h + 2 # same
    _pos = adim.getPosition() # fixme
    _hw = _layout_width//2
    _hh = _layout_height//2
    #
    # fill over the area where the text will be drawn
    #
    _box_gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
    _pixmap.draw_rectangle(_box_gc, gtk.TRUE, (_pdx - _hw), (_pdy - _hh),
                           _layout_width, _layout_height)
    del _box_gc # needed???
    #
    # draw the dimension text
    #
    _pgc = _da.get_style().black_gc
    if _dual_mode:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - _pri_h - 1), _pri_layout)
        _pixmap.draw_layout(_pgc, (_pdx - (_sec_w//2)), (_pdy + 1), _sec_layout)
        _pixmap.draw_line(_gc, (_pdx - _hw), _pdy, (_pdx + _hw), _pdy)
    else:
        _pixmap.draw_layout(_pgc, (_pdx - (_pri_w//2)), (_pdy - (_pri_h//2)), _pri_layout)

def angular_create(gtkimage, tool):
    assert len(tool) == 8, "Angular dimension tool should hold eight objects."
    y = tool.popObject()
    x = tool.popObject()
    p2 = tool.popObject()
    l2 = tool.popObject()
    p1 = tool.popObject()
    l1 = tool.popObject()
    vp = tool.popObject()
    vl = tool.popObject()
    ds = gtkimage.getOption('DIM_STYLE')
    try:
        adim = dimension.AngularDimension(vl, vp, l1, p1, l2, p2, x, y, ds)
    except:
        raise StandardError, "Unable to create AngularDimension!"
    set_dim_prefs(adim, gtkimage)
    set_adim_prefix_suffix(adim, gtkimage)
    try:
        gtkimage.addObject(adim)
    except StandardError, e:
        raise StandardError, "Failed to add Angular Dimension:" + str(e)
    gtkimage.setPrompt("Click on the angle vertex or an arc.")
    gtkimage.redraw()
    angular_mode_init(tool)

def _test_layer_arcs(gtkimage, lyr, x, y, tol):
    _adim = None
    for _arc in lyr.getLayerEntities("arc"):
        _arc_pt = _arc.mapPoint((x, y), tol)
        if _arc_pt is not None:
            _x, _y = _arc_pt.getCoords()
            _cp = _arc.getCenter()
            _ep1, _ep2 = _arc.getEndpoints()
            _p1 = lyr.hasPoint(_ep1)
            assert _p1 is not None, "Missing arc endpoint"
            _p2 = lyr.hasPoint(_ep2)
            assert _p2 is not None, "Missing arc endpoint"
            _ds = gtkimage.getOption("DIM_STYLE")
            _adim = dimension.AngularDimension(lyr, _cp, lyr, _p1, lyr, _p2,
                                               _x, _y, _ds)
            break
    return _adim

def angular_pts_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _pt_layer = _active_layer
    _pt = _active_layer.hasPoint((_x, _y), _tol)
    if _pt is None:
        _layers = []
        _layers.append(gtkimage.getTopLayer())
        while(len(_layers)):
            _layer = _layers.pop()
            if _layer is not _active_layer:
                _pt = _layer.hasPoint((_x, _y), _tol)
                if _pt is not None:
                    _pt_layer = _layer
                    break
            if _layer.getVisibility():
                for _child in _layer.getChildren():
                    _layers.append(_child)
    if _pt is not None:
        _x, _y = _pt.getCoords()
        tool.storeCoords(_x, _y)
        if len(tool) == 2:
            tool.pushObject(_pt_layer)
            tool.pushObject(_pt)
        else:
            _p1 = tool.popObject()
            _l1 = tool.popObject()
            _vp = tool.popObject()
            _vl = tool.popObject()
            _ds = gtkimage.getOption("DIM_STYLE")
            _adim = dimension.AngularDimension(_vl, _vp, _l1, _p1,
                                               _pt_layer, _pt,
                                               _x, _y, _ds)
            tool.pushObject(_adim)

def angular_second_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _test_pt = point.Point(_x, _y)
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _pt = _active_layer.hasPoint(_test_pt, _tol)
    _layer = _active_layer
    if _pt is None:
        _layers = []
        _layers.append(gtkimage.getTopLayer())
        while(len(_layers)):
            _layer = _layers.pop()
            if _layer is not _active_layer:
                _pt = _layer.hasPoint(_test_pt, _tol)
                if _pt is not None:
                    print "hit point in layer: " + _layer.getName()
                    _x, _y = _pt.getCoords()
                    tool.pushObject(_layer)
                    tool.pushObject(_pt)
                    tool.setLocation(_x, _y)
                    tool.storeCoords(_x, _y)
                    break
            if _layer.getVisibility():
                for _child in _layer.getChildren():
                    _layers.append(_child)
    if _pt is not None:
        _x, _y = _pt.getCoords()
        _count = len(tool)
        if _count == 2: # first pt/layer stored
            tool.pushObject(_layer)
            tool.pushObject(_pt)
            tool.setLocation(_x, _y)
            tool.storeCoords(_x, _y)
            gtkimage.setPrompt("Click on the second endpoint for the angular dimension.")
        elif _count == 4: # second pt/layer stored
            _p1 = tool.popObject()
            _l1 = tool.popObject()
            _vp = tool.popObject()
            _vl = tool.popObject()
            _ds = gtkimage.getOption("DIM_STYLE")
            _adim = dimension.AngularDimension(_vl, _vp, _l1, _p1,
                                               _layer, _pt,
                                               _x, _y, _ds)
            tool.pushObject(_adim)
            gtkimage.setPrompt("Click where the dimension text should be placed.")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)            
            tool.setHandler("motion_notify", adim_txt_motion_notify_cb)
            tool.setHandler("button_press", add_dimension)
        else:
            raise ValueError, "Unexpected tool length: %d" % _count

def angular_initial_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _test_pt = point.Point(_x, _y)
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _adim = None
    _pt = _active_layer.hasPoint(_test_pt, _tol)
    if _pt is None:
        print "no point in active layer"
        _adim = _test_layer_arcs(gtkimage, _active_layer, _x, _y, _tol)
        if _adim is not None:
            print "hit arc in active layer"
            tool.pushObject(_adim)
        else:
            print "no arc in active layer"
            _layers = []
            _layers.append(gtkimage.getTopLayer())
            while(len(_layers)):
                _layer = _layers.pop()
                if _layer is not _active_layer:
                    _pt = _layer.hasPoint(_test_pt, _tol)
                    if _pt is None:
                        _adim = _test_layer_arcs(gtkimage, _layer,
                                                 _x, _y, _tol)
                        if _adim is not None:
                            print "hit arc in layer: " + _layer.getName()
                            tool.pushObject(_adim)
                            break
                    else:
                        print "hit point in layer: " + _layer.getName()
                        _x, _y = _pt.getCoords()
                        tool.pushObject(_layer)
                        tool.pushObject(_pt)
                        tool.setLocation(_x, _y)
                        tool.storeCoords(_x, _y)
                        break
                if _layer.getVisibility():
                    for _child in _layer.getChildren():
                        _layers.append(_child)
    else:
        print "hit point in active layer"
        _x, _y = _pt.getCoords()
        tool.pushObject(_active_layer)
        tool.pushObject(_pt)
        tool.setLocation(_x, _y)
        tool.storeCoords(_x, _y)
    if _pt is None:
        if _adim is not None: # arc based angular dimension
            gtkimage.setPrompt("Click where the dimension text should be placed.")
            gtkimage.getGC().set_function(gtk.gdk.INVERT)            
            tool.setHandler("motion_notify", adim_txt_motion_notify_cb)
            tool.setHandler("button_press", add_dimension)
    else: # point based angular dimension
        tool.setHandler("button_press", angular_second_button_press_cb)
        gtkimage.setPrompt("Click on the first endpoint for the angular dimension.")
    
def angular_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", angular_initial_button_press_cb)
    tool.setHandler("initialize", angular_mode_init)
