#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# the GTK code for displaying a drawing
#

from __future__ import division

import math

import gtk
import gobject

import gtkmenus
import gtkdimension
import gtklayers
import gtkpycad

import Generic.image as image
import Generic.graphicobject as graphicobject
import Generic.point as point
import Generic.segment as segment
import Generic.circle as circle
import Generic.arc as arc
import Generic.hcline as hcline
import Generic.vcline as vcline
import Generic.acline as acline
import Generic.cline as cline
import Generic.ccircle as ccircle
import Generic.conobject as conobject
import Generic.color as color
import Generic.layer as layer
import Generic.segjoint as segjoint
import Generic.leader as leader
import Generic.dimension as dimension
import Generic.tools as tools
import Generic.globals as globals

def draw_layer(gtkimage, objlist, lyr):
    conobjs = []
    vobjs = []
    pts = []
    gc = gtkimage.getGC()
    gc.set_function(gtk.gdk.COPY)
    da = gtkimage.getDA()
    active_layer = gtkimage.getActiveLayer()
    if active_layer is lyr:
        active = True
    else:
        active = False
    window = da.window
    width, height = gtkimage.getSize()
    xmin, ymin, xmax, ymax = gtkimage.getView()
    _upp = gtkimage.getUnitsPerPixel()
    pixmap = gtkimage.getPixmap()
    draw_points = gtkimage.getOption('HIGHLIGHT_POINTS')
    inactive_layer_color = gtkimage.getOption('INACTIVE_LAYER_COLOR')
    for obj in objlist:
        if isinstance(obj, conobject._conobject):
            conobjs.append(obj)
        elif isinstance(obj, point.Point):
            pts.append(obj)
        else:
            vobjs.append(obj)
    #
    # do the construction lines/circles now. They all
    # have the same style bits, so the gc can be set
    # once and used for all the objects
    #
    
    gc.set_line_attributes(1, gtk.gdk.LINE_DOUBLE_DASH,
                           gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
    _dashlist = conobject.cline_linetype.getList()
    gc.set_dashes(0, _dashlist)
    if active:
        cline_color = conobject.cline_color
    else:
        cline_color = inactive_layer_color
    if gtkimage.hasGTKColor(cline_color):
        color = gtkimage.getGTKColor(cline_color)
    else:
        cstring = str(cline_color)
        color = da.get_colormap().alloc_color(cstring)
        gtkimage.saveGTKColor(cline_color, color)
    gc.set_foreground(color)
    gc.set_function(gtk.gdk.COPY)
    dtr = math.pi/180.0
    for cobj in conobjs:
        if isinstance(cobj, hcline.HCLine):
            _x, _y = cobj.getLocation().getCoords()
            _px, _py = gtkimage.coordToPixTransform(_x, _y)
            pixmap.draw_line(gc, 0, _py, width, _py)
        elif isinstance(cobj, vcline.VCLine):
            _x, _y = cobj.getLocation().getCoords()
            _px, _py = gtkimage.coordToPixTransform(_x, _y)
            pixmap.draw_line(gc, _px, 0, _px, height)
        elif isinstance(cobj, acline.ACLine):
            x, y = cobj.getLocation().getCoords()
            angle = cobj.getAngle()
            px, py = gtkimage.coordToPixTransform(x, y)
            if abs(abs(angle) - 90.0) < 1e-6:
                pixmap.draw_line(gc, px, 0, px, height)
            elif abs(angle) < 1e-6:
                pixmap.draw_line(gc, 0, py, width, py)
            else:
                x1 = None
                y1 = None
                x2 = None
                y2 = None
                m = math.tan(angle*dtr)
                b = y - (x * m)
                yt = (m * xmin) + b
                if (ymin - 1e-10) < yt < (ymax + 1e-10):
                    px, py = gtkimage.coordToPixTransform(xmin, yt)
                    x1 = px
                    y1 = py
                xt = (ymax - b)/m
                if (xmin - 1e-10) < xt < (xmax + 1e-10):
                    px, py = gtkimage.coordToPixTransform(xt, ymax)
                    if x1 is None:
                        x1 = px
                    elif x2 is None:
                        x2 = px
                    if y1 is None:
                        y1 = py
                    elif y2 is None:
                        y2 = py
                if x1 is None or x2 is None or y1 is None or y2 is None:
                    yt = (m * xmax) + b
                    if (ymin - 1e-10) < yt < (ymax + 1e-10):
                        px, py = gtkimage.coordToPixTransform(xmax, yt)
                        if x1 is None:
                            x1 = px
                        elif x2 is None:
                            x2 = px
                        if y1 is None:
                            y1 = py
                        elif y2 is None:
                            y2 = py
                if x1 is None or x2 is None or y1 is None or y2 is None:
                    xt = (ymin - b)/m
                    if (xmin - 1e-10) < xt < (xmax + 1e-10):
                        px, py = gtkimage.coordToPixTransform(xt, ymin)
                        if x1 is None:
                            x1 = px
                        elif x2 is None:
                            x2 = px
                        if y1 is None:
                            y1 = py
                        elif y2 is None:
                            y2 = py
                pixmap.draw_line(gc, x1, y1, x2, y2)
        elif isinstance(cobj, cline.CLine):
            p1, p2 = cobj.getKeypoints()
            p1x, p1y = p1.getCoords()
            p2x, p2y = p2.getCoords()
            if abs(p2x - p1x) < 1e-6: # vertical
                px, py = gtkimage.coordToPixTransform(p1x, p1y)                
                pixmap.draw_line(gc, px, 0, px, height)
            else:
                x1 = None
                y1 = None
                x2 = None
                y2 = None
                m = (p2y - p1y)/(p2x - p1x) # slope
                b = p1y - (p1x * m) # y-int
                yt = (m * xmin) + b
                if (ymin - 1e-10) < yt < (ymax + 1e-10):
                    px, py = gtkimage.coordToPixTransform(xmin, yt)
                    x1 = px
                    y1 = py
                xt = (ymax - b)/m
                if (xmin - 1e-10) < xt < (xmax + 1e-10):
                    px, py = gtkimage.coordToPixTransform(xt, ymax)
                    if x1 is None:
                        x1 = px
                    elif x2 is None:
                        x2 = px
                    if y1 is None:
                        y1 = py
                    elif y2 is None:
                        y2 = py
                if x1 is None or x2 is None or y1 is None or y2 is None:
                    yt = (m * xmax) + b
                    if (ymin - 1e-10) < yt < (ymax + 1e-10):
                        px, py = gtkimage.coordToPixTransform(xmax, yt)
                        if x1 is None:
                            x1 = px
                        elif x2 is None:
                            x2 = px
                        if y1 is None:
                            y1 = py
                        elif y2 is None:
                            y2 = py
                if x1 is None or x2 is None or y1 is None or y2 is None:
                    xt = (ymin - b)/m
                    if (xmin - 1e-10) < xt < (xmax + 1e-10):
                        px, py = gtkimage.coordToPixTransform(xt, ymin)
                        if x1 is None:
                            x1 = px
                        elif x2 is None:
                            x2 = px
                        if y1 is None:
                            y1 = py
                        elif y2 is None:
                            y2 = py
                pixmap.draw_line(gc, x1, y1, x2, y2)
        elif isinstance(cobj, ccircle.CCircle):
            _x, _y = cobj.getCenter().getCoords()
            _radius = cobj.getRadius()
            _px, _py = gtkimage.coordToPixTransform(_x, _y)
            _rx, _ry = gtkimage.coordToPixTransform((_x + _radius), _y)
            _pxrad = _rx - _px
            _pxmin = _px - _pxrad
            _pymin = _py - _pxrad
            _cw = _ch = _pxrad * 2
            pixmap.draw_arc(gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                            0, 360*64)
        else:
            pass
    if len(pts) and draw_points:
        _yellow = da.get_colormap().alloc_color("#ffff00")
        _cyan = da.get_colormap().alloc_color("#00ffff")
        gc.set_line_attributes(1, gtk.gdk.LINE_SOLID, gtk.gdk.CAP_BUTT,
                               gtk.gdk.JOIN_MITER)
        for _pt in pts:
            _objlist = lyr.usesObject(_pt)
            _gc_color = _yellow
            if len(_objlist) > 1:
                _nd = 0
                for _obj in _objlist:
                    if not isinstance(_obj, dimension.Dimension):
                        _nd = _nd + 1
                    if _nd > 1:
                        _gc_color = _cyan
                        break
            gc.set_foreground(_gc_color)
            _p1x, _p1y = gtkimage.coordToPixTransform(_pt.x, _pt.y)
            pixmap.draw_rectangle(gc, gtk.FALSE,
                                  (_p1x - 5), (_p1y - 5), 10, 10)
    for obj in vobjs:
        if isinstance(obj, graphicobject.GraphicObject):
            _dashlist = obj.getLinetype().getList()
            if _dashlist is not None:
                _linetype = gtk.gdk.LINE_DOUBLE_DASH
                gc.set_dashes(0, _dashlist)
            else:
                _linetype = gtk.gdk.LINE_SOLID
            if active:
                _c = obj.getColor()
            else:
                _c = inactive_layer_color
            if gtkimage.hasGTKColor(_c):
                color = gtkimage.getGTKColor(_c)
            else:
                cstring = str(_c)
                color = da.get_colormap().alloc_color(cstring)
                gtkimage.saveGTKColor(_c, color)
            # _t = obj.getThickness()
            _thickness = int(obj.getThickness()/_upp)
            if _thickness == 0:
                _thickness = 1
            gc.set_foreground(color)                
            gc.set_line_attributes(_thickness, _linetype,
                                   gtk.gdk.CAP_ROUND, gtk.gdk.JOIN_MITER)
        if isinstance(obj, segment.Segment):
            _p1, _p2 = obj.getEndpoints()
            _p1x, _p1y = gtkimage.coordToPixTransform(_p1.x, _p1.y)
            _p2x, _p2y = gtkimage.coordToPixTransform(_p2.x, _p2.y)
            pixmap.draw_line(gc, _p1x, _p1y, _p2x, _p2y)
        elif isinstance(obj, arc.Arc):
            _x, _y = obj.getCenter().getCoords()
            _r = obj.getRadius()
            _px, _py = gtkimage.coordToPixTransform(_x, _y)
            _rx, _ry = gtkimage.coordToPixTransform((_x + _r), _y)
            _radius = _rx - _px
            _pxmin = _px - _radius
            _pymin = _py - _radius
            _cw = _ch = _radius * 2
            _sa = obj.getStartAngle()
            _ea = obj.getEndAngle()
            if abs(_sa - _ea) < 1e-10:
                pixmap.draw_arc(gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                                0, (360*64))
            elif _sa > _ea:
                pixmap.draw_arc(gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                                (_sa*64), (360.0-_sa)*64)
                pixmap.draw_arc(gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                                0, (_ea*64))
            else:
                _angle = _ea - _sa
                pixmap.draw_arc(gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                                (_sa*64), (_angle*64))
        elif isinstance(obj, circle.Circle):
            _x, _y = obj.getCenter().getCoords()
            _r = obj.getRadius()
            _px, _py = gtkimage.coordToPixTransform(_x, _y)
            _rx, _ry = gtkimage.coordToPixTransform((_x + _r), _y)
            _radius = _rx - _px
            _pxmin = _px - _radius
            _pymin = _py - _radius
            _cw = _ch = _radius * 2
            pixmap.draw_arc(gc, gtk.FALSE, _pxmin, _pymin, _cw, _ch,
                            0, 360*64)
        elif isinstance(obj, segjoint.Chamfer):
            _p1, _p2 = obj.getMovingPoints()
            _p1x, _p1y = gtkimage.coordToPixTransform(_p1.x, _p1.y)
            _p2x, _p2y = gtkimage.coordToPixTransform(_p2.x, _p2.y)
            pixmap.draw_line(gc, _p1x, _p1y, _p2x, _p2y)
        elif isinstance(obj, segjoint.Fillet):
            cx, cy = obj.getCenter()
            pcx, pcy = gtkimage.coordToPixTransform(cx, cy)
            p1, p2 = obj.getMovingPoints()
            p1x, p1y = gtkimage.coordToPixTransform(p1.x, p1.y)
            p2x, p2y = gtkimage.coordToPixTransform(p2.x, p2.y)
            r = obj.getRadius()
            rx, ry = gtkimage.coordToPixTransform((cx + r), cy)
            radius = rx - pcx
            print "radius: %d" % radius
            pxmin = pcx - radius
            pymin = pcy - radius
            cw = ch = radius * 2
            sa1, sa2 = obj.getAngles()
            amin = min(sa1, sa2)
            amax = max(sa1, sa2)
            if amax - amin > 180.0:
                a1 = amax
                a2 = amin
            else:
                a1 = amin
                a2 = amax
            print "a1: %g" % a1
            print "a2: %g" % a2
            if a1 > a2:
                pixmap.draw_arc(gc, gtk.FALSE, pxmin, pymin, cw, ch,
                                (a1*64), (360.0-a1)*64)
                pixmap.draw_arc(gc, gtk.FALSE, pxmin, pymin, cw, ch,
                                0, (a2*64))
            else:
                angle = a2 - a1
                pixmap.draw_arc(gc, gtk.FALSE, pxmin, pymin, cw, ch,
                                (a1*64), (angle*64))
        elif isinstance(obj, leader.Leader):
            _p1, _p2, _p3 = obj.getLeaderPoints()
            _p1x, _p1y = gtkimage.coordToPixTransform(_p1.x, _p1.y)
            _p2x, _p2y = gtkimage.coordToPixTransform(_p2.x, _p2.y)
            _p3x, _p3y = gtkimage.coordToPixTransform(_p3.x, _p3.y)
            pixmap.draw_line(gc, _p1x, _p1y, _p2x, _p2y)
            pixmap.draw_line(gc, _p2x, _p2y, _p3x, _p3y)
            _apts = []
            _apts.append((_p3x, _p3y))
            _pts = obj.getArrowPoints()
            _a1x, _a1y = gtkimage.coordToPixTransform(_pts[0], _pts[1])
            _apts.append((_a1x, _a1y))
            _a2x, _a2y = gtkimage.coordToPixTransform(_pts[2], _pts[3])
            _apts.append((_a2x, _a2y))
            pixmap.draw_polygon(gc, gtk.TRUE, _apts)
        elif isinstance(obj, dimension.HorizontalDimension):
            gtkdimension.draw_linear_dimension(gtkimage, obj, active)
        elif isinstance(obj, dimension.VerticalDimension):
            gtkdimension.draw_linear_dimension(gtkimage, obj, active)
        elif isinstance(obj, dimension.LinearDimension):
            gtkdimension.draw_linear_dimension(gtkimage, obj, active)
        elif isinstance(obj, dimension.RadialDimension):
            gtkdimension.draw_radial_dimension(gtkimage, obj, active)
        elif isinstance(obj, dimension.AngularDimension):
            gtkdimension.draw_angular_dimension(gtkimage, obj, active)
        else:
            print "skipping obj: " + `obj`

    #
    # reset line thickness to 1
    #
    
    gc.set_line_attributes(1, gtk.gdk.LINE_DOUBLE_DASH,
                           gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
    
def da_expose_event(widget, event, gtkimage):
    _pixmap = gtkimage.getPixmap()
    _x, _y, _width, _height = event.area
    _gc = widget.get_style().fg_gc[gtk.STATE_NORMAL]
    # _gc.set_function(gtk.gdk.COPY)
    widget.window.draw_drawable(_gc, _pixmap, _x, _y, _x, _y, _width, _height)
    return gtk.TRUE

def da_realize_event(widget, gtkimage):
    _win = widget.window
    _width, _height = _win.get_size()
    gtkimage.setSize(_width, _height)
    widget.set_size_request(100,100)
    _gc = _win.new_gc()
    _gc.set_exposures(gtk.TRUE)
    gtkimage.setGC(_gc)    

def da_configure_event(widget, event, gtkimage):
    _win = widget.window
    _width, _height = _win.get_size()
    _disp_width, _disp_height = gtkimage.getSize()
    if _disp_width != _width or _disp_height != _height:
        gtkimage.setSize(_width, _height)
        _pixmap = gtk.gdk.Pixmap(_win, _width, _height)
        _gc = widget.get_style().fg_gc[gtk.STATE_NORMAL]
        _pixmap.draw_rectangle(_gc, gtk.TRUE, 0, 0, _width, _height)
        gtkimage.setPixmap(_pixmap)
        _xmin, _ymin = gtkimage.getView()[:2]
        if _xmin is None or _ymin is None:
            _xmin = 1.0
            _ymin = 1.0
        _upp = gtkimage.getUnitsPerPixel()
        gtkimage.setView(_xmin, _ymin, _upp)
        gtkimage.redraw()
    return gtk.TRUE

def set_toolpoint(gtkimage, event):
    _x = event.x
    _y = event.y
    _tx, _ty = gtkimage.pixToCoordTransform(_x, _y)
    gtkimage.setPoint(_tx, _ty)

def da_general_event(widget, event, gtkimage):
    type = event.type
    tool = gtkimage.getTool()
    if type == gtk.gdk.BUTTON_PRESS:
        set_toolpoint(gtkimage, event)
        button = event.button
        if button == 1 and tool.hasHandler("button_press"):
            handler = tool.getHandler("button_press")
            handler(gtkimage, widget, event, tool)
    elif type == gtk.gdk.BUTTON_RELEASE:
        set_toolpoint(gtkimage, event)
        button = event.button
        if button == 1 and tool.hasHandler("button_release_event"):
            handler = tool.getHandler("button_release_event")
            handler(gtkimage, widget, event, tool)
    elif type == gtk.gdk.MOTION_NOTIFY:
        set_toolpoint(gtkimage, event)        
        if tool.hasHandler("motion_notify"):
            handler = tool.getHandler("motion_notify")
            handler(gtkimage, widget, event, tool)
#     elif type == gtk.gdk.KEY_PRESS:
#         # print "Got key press:"
#         key = event.keyval
#         if key == gtk.keysyms.Page_Up or key == gtk.keysyms.Page_Down \
#            or key == gtk.keysyms.Left or key == gtk.keysyms.Right \
#            or key == gtk.keysyms.Up or key == gtk.keysyms.Down:
#             print "Got special key"
#         handler = tool.getHandler("keypress_event")
#         if handler is not None:
#             handler(dw, widget, event, tool)
    elif type == gtk.gdk.ENTER_NOTIFY:
        set_toolpoint(gtkimage, event)
    elif type == gtk.gdk.LEAVE_NOTIFY:
        set_toolpoint(gtkimage, event)
    else:
        # print "Got type %d" % type
        pass
    return gtk.FALSE

def entry_event(widget, gtkimage):
    tool = gtkimage.getTool()
    if tool.hasHandler("entry_event"):
        handler = tool.getHandler("entry_event")
        try:
            handler(gtkimage, widget, tool)
        except: # need better error handling ...
            pass
    else:
        gtkimage.entry.delete_text(0,-1)
    return gtk.FALSE

def close_image(widget, gtkimage):
    """Close a window containing a GTKImage.

close_image(gtkimage)    
    """
    gtkimage.close()
    gtkpycad.del_image(gtkimage)
    return gtk.FALSE

def tree_view_button_cb(widget, event, gtkimage):
    retval = gtk.FALSE
    button = event.button
    x = event.x
    y = event.y
    if event.window is widget.get_bin_window():
        model = widget.get_model()
        path, col, cx, cy = widget.get_path_at_pos(x, y)
        iter = model.get_iter(path)
        lyr = model.get_value(iter, 1)
        if button == 3:
            menu = gtklayers.make_layer_menu(gtkimage, lyr)
            menu.popup(None, None, None, button, event.time)
            retval = gtk.TRUE
    return retval

def reordered_cb(model, path, iter, new_order):
    print "in reordered_cb()"
    print "model: " + `model`
    print "path: " + `path`
    print "iter: " + `iter`
    print "new_order: " + `new_order`

# def columns_changed_cb(tree_view, col, prev_col, next_col, gtkimage):
def columns_changed_cb(tree_view, gtkimage):
    print "in columns_changed_cb()"
    print "tree_view: " + `tree_view`
    # print "col: " + `col`
    # print "prev_col: " + `prev_col`
    # print "next_col: " + `next_col`
    # return gtk.FALSE

def cell_edited_cb(cell_renderer, path, text, model):
    iter = model.get_iter_from_string(path)
    lyr = model.get_value(iter, 1)
    lyr.setName(text)
    model.set(iter, 0, text)

def tree_select_cb(selection, gtkimage):
    if selection is not None:
        model, iter = selection.get_selected()
        if iter is not None:
            lyr = model.get_value(iter, 1)
            gtkimage.setActiveLayer(lyr)

def model_find_layer(model, lyr, iter=None):
    if iter is None:
        iter = model.get_iter_first()
    path = None
    mlyr = model.get_value(iter, 1)
    print "model layer: " + mlyr.getName()
    if mlyr is lyr:
        path = model.get_path(iter)
    else:
        if model.iter_has_child(iter):
            child = model.iter_children(iter)
            while child is not None:
                path = model_find_layer(model, lyr, child)
                if path is not None:
                    break
                child = model.iter_next(child)
    return path

def model_add_layer(model, path, iter, lyr):
    parent = lyr.getParent()
    mlayer = model.get_value(iter, 1)
    val = gtk.FALSE
    if mlayer is parent:
        iter = model.append(iter)
        model.set(iter, 0, lyr.getName())
        model.set(iter, 1, lyr)
        val = gtk.TRUE
    return val

def model_del_layer(model, path, iter, lyr):
    mlayer = model.get_value(iter, 1)
    val = gtk.FALSE
    if mlayer is lyr:
        model.remove(iter)
        val = gtk.TRUE
    return val

def model_rename_layer(model, path, iter, lyr):
    mlayer = model.get_value(iter, 1)
    val = gtk.FALSE
    if mlayer is lyr:
        model.set(iter, 0, lyr.getName())
        val = gtk.TRUE
    return val

class GTKImage(image.Image):
    """The GTK wrapping around an Image
    """
    def __init__(self):
        image.Image.__init__(self)
        self.__window = gtk.Window()
        self.__window.set_title("Untitled")
        self.__window.connect("destroy", close_image, self)
        width = min(1024, int(0.8 * float(gtk.gdk.screen_width())))
        height = min(768, int(0.8 * float(gtk.gdk.screen_height())))
        self.__window.set_default_size(width,height)

        main_vbox = gtk.VBox(gtk.FALSE, 2)
        main_vbox.set_border_width(2)
        self.window.add(main_vbox)

        #
        # accelerators
        #

        self.__accel = gtk.AccelGroup()
        self.__window.add_accel_group(self.__accel)
        
        #
        # menu bar
        #
        
        self.__mb = gtk.MenuBar()
        main_vbox.pack_start(self.__mb, gtk.FALSE, gtk.FALSE)

        gtkmenus.fill_menubar(self.__mb, self)

        #
        # drawing window has Horizontal Pane:
        # left side: stuff for layer display
        # right side: drawing area
        #

        pane = gtk.HPaned()
        main_vbox.pack_start(pane)

        frame1 = gtk.Frame()
        pane.pack1(frame1, gtk.TRUE, gtk.FALSE)
        pane.set_position(100)

        #
        # layer display stuff
        #

        self.__sw = gtk.ScrolledWindow()
        self.__sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)

        self.__model = gtk.TreeStore(gobject.TYPE_STRING,
                                     gobject.TYPE_PYOBJECT)
        self.__model.connect("rows-reordered", reordered_cb)
        iter = self.__model.append(None)
        top_layer = self.getTopLayer()
        self.__model.set(iter, 0, top_layer.getName())
        self.__model.set(iter, 1, top_layer)
        self.__tree_view = gtk.TreeView(self.__model)
        self.__tree_view.set_reorderable(gtk.TRUE) # drag-and-drop
        # self.__tree_view.connect("columns-changed",
                                 # columns_changed_cb,
                                 # self)
        self.__tree_view.connect("button_press_event",
                                 tree_view_button_cb, self)
        # self.__tree_view.set_headers_visible(gtk.FALSE)
        # self.__tree_view.expand_all()
        select = self.__tree_view.get_selection()
        select.set_mode(gtk.SELECTION_SINGLE)
        select.connect("changed", tree_select_cb, self)
        renderer = gtk.CellRendererText()
        # renderer.set_property("editable", gtk.TRUE)
        renderer.connect("edited", cell_edited_cb, self.__model)
        column = gtk.TreeViewColumn("Layers", renderer, text=0)
        self.__tree_view.append_column(column)
    
        self.__sw.add(self.__tree_view)
        frame1.add(self.__sw)

        #
        # drawing area
        #

        self.__disp_width = None
        self.__disp_height = None
        self.__units_per_pixel = 1.0
        self.__da = gtk.DrawingArea()

        black = gtk.gdk.color_parse('black')
        self.__da.modify_fg(gtk.STATE_NORMAL, black)
        self.__da.modify_bg(gtk.STATE_NORMAL, black)
        pane.pack2(self.__da, gtk.TRUE, gtk.FALSE)
        self.__da.set_flags(gtk.CAN_FOCUS)
        
        self.__da.connect("event", da_general_event, self)
        self.__da.connect("expose_event", da_expose_event, self)
        self.__da.connect("realize", da_realize_event, self)
        self.__da.connect("configure_event", da_configure_event, self)

        self.__da.set_events(gtk.gdk.EXPOSURE_MASK |
                             gtk.gdk.LEAVE_NOTIFY_MASK |
                             gtk.gdk.BUTTON_PRESS_MASK |
                             gtk.gdk.BUTTON_RELEASE_MASK |
                             gtk.gdk.ENTER_NOTIFY_MASK|
                             gtk.gdk.LEAVE_NOTIFY_MASK|
                             gtk.gdk.KEY_PRESS_MASK |
                             gtk.gdk.KEY_RELEASE_MASK |
                             gtk.gdk.POINTER_MOTION_MASK)

        lower_hbox = gtk.HBox(gtk.FALSE, 2)
        main_vbox.pack_start(lower_hbox, gtk.FALSE, gtk.FALSE)

        self.__prompt = gtk.Label("Enter Command:")
        lower_hbox.pack_start(self.__prompt, gtk.FALSE, gtk.FALSE)

        #
        # where the action is taking place
        #
        
        self.__coords = gtk.Label('(0,0)')
        lower_hbox.pack_end(self.__coords, gtk.FALSE, gtk.FALSE)
        
        self.__point_x = 0.0
        self.__point_y = 0.0

        #
        # command entry area
        #

        self.__entry = gtk.Entry()
        main_vbox.pack_start(self.__entry, gtk.FALSE, gtk.FALSE)
        self.entry.connect("activate", entry_event, self)

        #
        # the Pixmap and GraphicContext for the drawing
        #
        
        self.__pixmap = None
        self.__gc = None

        #
        # our tool used to draw stuff
        #

        self.__tool = tools.Tool()

        #
        # the viewable region and tolerance in the drawing
        #

        self.__xmin = None
        self.__ymin = None
        self.__xmax = None
        self.__ymax = None
        self.__tolerance = 1e-10

        #
        # allocated color and linetype storage
        #

        self.__gtklinetypes = {}
        self.__gtkcolors = {}

        #
        # set the background color
        #
        
        self.setOption('BACKGROUND_COLOR', color.Color(0,0,0))
        
    def close(self):
        """Release the entites stored in the drawing.

close()        
        """
        image.Image.close(self)
        self.__window.destroy()
        self.__gtkcolors.clear()
        self.__gtklinetypes.clear()
        
    def getAccel(self):
        """Return the gtk.AccelGroup in the GTKImage.

getAccel()        
        """
        return self.__accel

    accel = property(getAccel, None, None, "Accel group in the GTKImage.")
    
    def getWindow(self):
        """Return the gtk.Window in the GTKImage.

getWindow()        
        """
        return self.__window

    window = property(getWindow, None, None, "Main GTK Window for a GTKImage.")
    
    def getEntry(self):
        """Return the gtk.Entry in the GTKImage.

getEntry()        
        """
        return self.__entry

    entry = property(getEntry, None, None, "Entry box for a GTKImage.")
    
    def getDA(self):
        """Return the gtk.DrawingArea in the GTKImage.

getDA()
        """
        return self.__da

    da = property(getDA, None, None, "DrawingArea for a GTKImage.")

    def getPixmap(self):
        """Return the Pixmap for the GTKImage.

getPixmap()        
        """
        return self.__pixmap

    def setPixmap(self, pixmap):
        """Set the Pixmap for the GTKImage.

setPixmap(pixmap)
        """
        self.__pixmap = pixmap
        
    pixmap = property(getPixmap, setPixmap, None, "Pixmap for a GTKImage.")

    def getPrompt(self):
        """Return the current prompt string.

getPrompt()        
        """
        return self.__prompt.get_label()
    
    def setPrompt(self, prompt):
        """Set the current prompt string.

setPrompt(prompt)
        """
        if not isinstance(prompt, (str, unicode)):
            raise TypeError, "Invalid prompt: " + `prompt`
        self.__prompt.set_text(prompt)

    prompt = property(getPrompt, setPrompt, None, "Prompt string.")

    def getTool(self):
        """Return the tool used in this drawing.

getTool()        
        """
        return self.__tool

    tool = property(getTool, None, None, "Tool for adding/modifying entities.")

    def setOption(self, key, value):
        """

setOption(key, value)

The "key" must be a string, and "value" can be anything.
Using the same key twice will result on the second value
overwriting the first.

This method extends the image.Image.setOption() method and
handles options that are related to the graphical appearance
of the image.
        """
        image.Image.setOption(self, key, value)
        if key == 'BACKGROUND_COLOR':
            bcol = str(value)
            col = gtk.gdk.color_parse(bcol)
            self.__da.modify_fg(gtk.STATE_NORMAL, col)
            self.__da.modify_bg(gtk.STATE_NORMAL, col)
            self.redraw()

    def getUnitsPerPixel(self):
        """Return the current value of units/pixel.

getUnitsPerPixel()        
        """
        return self.__units_per_pixel

    def setUnitsPerPixel(self, upp):
        """Set the current value of units/pixel.

setUnitsPerPixel(upp)

The argument "upp" should be a positive float value.
        """
        _upp = upp
        if not isinstance(_upp, float):
            _upp = float(upp)
        if _upp < 1e-10:
            raise ValueError, "Invalid scale value: " + `_upp`
        self.__units_per_pixel = _upp
        self.__tolerance = _upp * 5.0
    
    def setView(self, xmin, ymin, scale=None):
        """Set the current visible area in a drawing.

setView(xmin, ymin[, scale])

xmin: Minimum visible x-coordinate
ymin: Minimum visible y-coordinate

The optional argument "scale" defaults to the current
value of units/pixel (set the {get/set}UnitsPerPixel()
method. This value must be a positive float.
        """
        _xmin = xmin
        if not isinstance(_xmin, float):
            _xmin = float(xmin)
        _ymin = ymin
        if not isinstance(_ymin, float):
            _ymin = float(ymin)
        _scale = scale
        if _scale is None:
            _scale = self.__units_per_pixel
        if not isinstance(_scale, float):
            _scale = float(scale)
        if _scale < 1e-10:
            raise ValueError, "Invalid scale value: " + `_scale`
        _xmax = _xmin + (_scale * self.__disp_width)
        _ymax = _ymin + (_scale * self.__disp_height)
        if _scale != self.__units_per_pixel:
            self.__units_per_pixel = _scale
        self.__tolerance = self.__units_per_pixel * 5.0
        self.__xmin = _xmin
        self.__ymin = _ymin
        self.__xmax = _xmax
        self.__ymax = _ymax

    def getView(self):
        """Return the current visible area in a drawing.

getView()

This method returns a tuple with four float values:

(xmin, ymin, xmax, ymax)

If the view has never been set, each of these values
will be None.
        """
        return (self.__xmin, self.__ymin, self.__xmax, self.__ymax)

    view = property(getView, setView, None, "The visible area in a drawing.")

    def setActiveLayer(self, l=None):
        """Set the active Layer in the Image.

setActiveLayer(l)

This method extends the image::setActiveLayer() method.
        """
        try:
            image.Image.setActiveLayer(self, l)
            self.redraw()
        except: # need better error handling
            return
        model = self.__model
        path = model_find_layer(model, l)
        assert path is not None, "Didn't find Layer in model!"
        ppath = path[:-1]
        while len(ppath):
            if not self.__tree_view.row_expanded(ppath):
                self.__tree_view.expand_row(ppath, gtk.FALSE)
            ppath = ppath[:-1]
        select = self.__tree_view.get_selection()
        select.select_path(path)

    def addLayer(self, l):
        """Add a new Layer as a Child of the active layer's parent.

addLayer(l)

This method extends the image::addLayer() method.
        """
        try:
            image.Image.addLayer(self, l)
        except: # need better error handling
            return
        model = self.__model
        model.foreach(model_add_layer, l)
        path = model_find_layer(model, l)
        assert path is not None, "Didn't find Layer in model!"
        ppath = path[:-1]
        while len(ppath):
            if not self.__tree_view.row_expanded(ppath):
                self.__tree_view.expand_row(ppath, gtk.FALSE)
            ppath = ppath[:-1]
        select = self.__tree_view.get_selection()
        select.select_path(path)

    def addChildLayer(self, l, p=None):
        """Add a new Layer as a child to another layer.

addChildLayer(l [, p])

There is a one required argument:

l: The child Layer

There is one optional argument:

p: The new parent Layer of the child

The default parent is the currently active layer.

This method extends the image::addChildLayer() method.
        """
        try:
            image.Image.addChildLayer(self, l, p)
        except: # need better error handling ...
            return
        model = self.__model
        model.foreach(model_add_layer, l)
        path = model_find_layer(model, l)
        assert path is not None, "Didn't find Layer in model!"
        ppath = path[:-1]
        while len(ppath):
            if not self.__tree_view.row_expanded(ppath):
                self.__tree_view.expand_row(ppath, gtk.FALSE)
            ppath = ppath[:-1]
        select = self.__tree_view.get_selection()
        select.select_path(path)

    def delLayer(self, l):
        """Remove a Layer from the drawing.

delLayer(l)

This method extends the image::delLayer() method.
        """
        try:
            image.Image.delLayer(self, l)
        except:
            return
        model = self.__model
        path = model_find_layer(model, l)
        assert path is not None, "Didn't find Layer in model!"
        ppath = path[:-1]
        while len(ppath):
            if not self.__tree_view.row_expanded(ppath):
                self.__tree_view.expand_row(ppath, gtk.FALSE)
            ppath = ppath[:-1]
        select = self.__tree_view.get_selection()
        select.select_path(path[:-1])
        model.foreach(model_del_layer, l)

    def renameLayer(self, l):
        """Rename a Layer in the drawing.

renameLayer(l)

This method is used to update the Layer display
        """
        if not isinstance(l, layer.Layer):
            raise TypeError, "Invalid layer: " + `l`
        self.__model.foreach(model_rename_layer, l)

    def addObject(self, obj, l=None):
        """Add an object to the Drawing.

addObject(obj [, l])

This method extends the image::addObject() method.
        """
        try:
            image.Image.addObject(self, obj, l)
        except StandardError, e:
            print "Failed to add object: " + `obj` + str(e)
        #
        # it may be good to call redraw() here instead of
        # the multitude of places where it is called now ...
        #
        # self.redraw()

    def getTolerance(self):
        """Return the current drawing tolerance.

getTolerance()
        """
        return self.__tolerance

    tolerance = property(getTolerance, None, None, "Drawing tolerance.")
    
    def getGC(self):
        """Return the GraphicContext allocated for the GTKImage.

getGC()        
        """
        return self.__gc

    def setGC(self, gc):
        """Set the GraphicContext for the GTKImage.

setGC(gc)        
        """
        if not isinstance(gc, gtk.gdk.GC):
            raise TypeError, "Invalid GC object: " + `gc`
        if self.__gc is None:
            self.__gc = gc
            
    gc = property(getGC, None, None, "GraphicContext for the GTKImage.")
    
    def getPoint(self):
        """Get the current point where the tool is located.

getPoint()

This function returns a tuple with two floats

(x,y)

x: x-coordinate
y: y-coordinate
        """
        return (self.__point_x, self.__point_y)

    def setPoint(self, x, y):
        """Store the point where the tool currently is located at.

setPoint(x,y)

The x and y arguments should be floats.
        """
        _x = x
        _y = y
        try:
            if not isinstance(_x, float):
                _x = float(x)
            if not isinstance(_y, float):
                _y = float(y)
            cstring = "%.4f, %.4f" % (_x,_y)
            self.__coords.set_text(cstring)
            self.__point_x = _x
            self.__point_y = _y
        except: # need better error handling
            pass

    point = property(getPoint, setPoint, None, "Current tool location.")

    def getSize(self):
        """Return the size of the DrawingArea window.

getSize()        
        """
        return (self.__disp_width, self.__disp_height)
    
    def setSize(self, width, height):
        """Set the size of the DrawingArea window.

setSize(width, height)
        """
        if width < 0:
            raise ValueError, "Invalid drawing area width: " + `width`
        if height < 0:
            raise ValueError, "Invalid drawing area height: " + `height`
        self.__disp_width = width
        self.__disp_height = height

    def pixToCoordTransform(self, xp, yp):
        """Convert from pixel coordinates to x-y coordinates.

pixToCoordTransform(xp, yp)

The function arguments are:

xp: pixel x value
yp: pixel y value

The function returns a tuple holding two float values
        """
        upp = self.__units_per_pixel
        xc = self.__xmin + (xp * upp)
        yc = self.__ymax - (yp * upp)
        return (xc, yc)

    def coordToPixTransform(self, xc, yc):
        """Convert from x-y coordinates to pixel coordinates

coordToPixTransform(xc, yc)

The function arguments are:

xc: x coordinate
yp: y coordinate

The function returns a tuple holding two integer values
        """
        upp = self.__units_per_pixel
        xp = int((xc - self.__xmin)/upp)
        yp = int((self.__ymax - yc)/upp)
        return (xp, yp)

    def saveGTKColor(self, key, val):
        """Store an allocated color

saveGTKColor(key, val)

The arguments are:
key: A Color object
val: The allocated color returned from the GTK call
        """
        if not isinstance(key, color.Color):
            raise TypeError, "Invalid Color object: " + `key`
        self.__gtkcolors[key] = val

    def getGTKColor(self, col):
        """

getGTKColor(col)

The argument color should be a Color object.
        """
        val = None
        if self.__gtkcolors.has_key(col):
            val = self.__gtkcolors[col]
        return val

    def hasGTKColor(self, col):
        """

hasGTKColor(col)

The argument color should be a Color object.
        """
        if not isinstance(col, color.Color):
            raise TypeError, "Invalid Color object: " + `col`
        return col in self.__gtkcolors

    def fitImage(self):
        """Redraw the image so all entities are visible in the window.

fitImage()        
        """
        fw = float(self.__disp_width)
        fh = float(self.__disp_height)
        sr = fw/fh
        xmin, ymin, xmax, ymax = self.getExtents()
        xdiff = abs(xmax - xmin)
        ydiff = abs(ymax - ymin)
        xmid = (xmin + xmax)/2.0
        ymid = (ymin + ymax)/2.0
        xs = xdiff/fw
        ys = ydiff/fh
        if xs > ys:
            scale = xs * 1.05 # make it a little larger
        else:
            scale = ys * 1.05 # make it a little larger
        xm = xmid - (fw/2.0) * scale
        ym = ymid - (fh/2.0) * scale
        self.setView(xm, ym, scale)
        self.redraw()

    def refresh(self):
        """This method does a screen refresh.

refresh()

If entities in the drawing have been added, removed, or
modified, use the redraw() method.
        """
        _da = self.__da
        flags = _da.flags()
        if (flags & (1 << 7)): # drawing area is mapped
            _gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
            _gc.set_function(gtk.gdk.COPY)
            _da.queue_draw()
            
    def redraw(self):
        """This method draws all the objects visible in the window.

redraw()
        """
        _da = self.__da
        _flags = _da.flags()
        if (_flags & (1 << 7)): # drawing area is mapped                
            _xmin = self.__xmin
            _ymin = self.__ymin
            _xmax = self.__xmax
            _ymax = self.__ymax
            _gc = _da.get_style().fg_gc[gtk.STATE_NORMAL]
            self.__pixmap.draw_rectangle(_gc, gtk.TRUE, 0, 0,
                                         self.__disp_width, self.__disp_height)
            _active_layer = self.getActiveLayer()
            _layers = []            
            _layers.append(self.getTopLayer())
            while (len(_layers)):
                _layer = _layers.pop()
                if _layer is not _active_layer and _layer.getVisibility():
                    _objs = _layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
                    draw_layer(self, _objs, _layer)
                if _layer.hasChildren():
                    for _child in _layer.getChildren():
                        _layers.append(_child)
            if _active_layer.getVisibility():
                _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
                draw_layer(self, _objs, _active_layer)
            _da.queue_draw() # generate an expose event
