#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# GTK Menu building code
#

import os
import sys

import gtk
import gtk.keysyms

import gtkimage
import gtkedit
import gtkentities
import gtkconobjs
import gtkprefs
import gtkdimprefs
import gtkmodify
import gtklayers
import gtkdimension
import gtkpycad
import Generic.globals as globals
import Generic.fileio as fileio
import Generic.imageio as imageio
import Generic.menus

#
# dictionary used to map strings to stock icons
#

iconmap = {
    'NEW' : gtk.STOCK_NEW,
    'QUIT' : gtk.STOCK_QUIT,
    'OPEN' : gtk.STOCK_OPEN,
    'CLOSE': gtk.STOCK_CLOSE,
    'DELETE' : gtk.STOCK_DELETE,
    'SAVE' : gtk.STOCK_SAVE,
    'SAVE_AS' : gtk.STOCK_SAVE_AS,
    'UNDO' : gtk.STOCK_UNDO,
    'CUT' : gtk.STOCK_CUT,
    'COPY' : gtk.STOCK_COPY,
    'PASTE': gtk.STOCK_PASTE,
    'PREFS' : gtk.STOCK_PREFERENCES,
    'ZOOM_IN' : gtk.STOCK_ZOOM_IN,
    'ZOOM_OUT' : gtk.STOCK_ZOOM_OUT,
    'ZOOM_FIT' : gtk.STOCK_ZOOM_FIT,
    }

#
# dictionary used to map letters to keysyms
#

accelmap = {
    'N' : gtk.keysyms.N,
    'O' : gtk.keysyms.O,
    'W' : gtk.keysyms.W,
    'S' : gtk.keysyms.S,
    'Q' : gtk.keysyms.Q,
    'Z' : gtk.keysyms.Z,
    'X' : gtk.keysyms.X,
    'C' : gtk.keysyms.C,
    'P' : gtk.keysyms.P,
    'V' : gtk.keysyms.V
    }

#
# callbacks for the menu items
#

def file_menu_init_cb(menuitem, gtkimage):
    # print "in file_menu_init_cb()"
    return gtk.FALSE

def file_new_cb(menuitem, gtkimg):
    img = gtkimage.GTKImage()
    gtkpycad.add_image(img)
    img.window.show_all()
    return gtk.FALSE

def file_open_cb(menuitem, gtkimg):
    open_flag = False
    fname = None
    dialog = gtk.FileSelection("Open File ...")
    dialog.set_transient_for(gtkimg.getWindow())
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        fname = dialog.get_filename()
        open_flag = True
    dialog.destroy()
    if open_flag:
        img = gtkimage.GTKImage()
        try:
            imfile = fileio.CompFile(fname, "r")
            try:
                imageio.load_image(img, imfile)
            finally:
                imfile.close()
        except (IOError, OSError), e:
            print "Can't open '%s': %s" % (fname, e)
            return gtk.FALSE
        except StandardError, e:
            print "Error: " + str(e)
            return gtk.FALSE
        img.setFilename(fname)
        img.getWindow().set_title(os.path.basename(fname))
        gtkpycad.add_image(img)
        img.window.show_all()
        img.fitImage()
    return gtk.FALSE

def file_close_cb(menuitem, gtkimage):
    gtkpycad.del_image(gtkimage)
    return gtk.FALSE

def file_save_cb(menuitem, gtkimage):
    fname = gtkimage.getFilename()
    if fname is None:
        window = gtkimage.getWindow()
        fname = window.get_title()
        dialog = gtk.FileSelection("Save As ...")
        dialog.set_transient_for(gtkimage.getWindow())
        dialog.set_filename(fname)
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            fname = dialog.get_filename()
            if fname == "":
                fname = 'Untitled.xml'
            if not fname.endswith('.xml'):
                fname = fname + '.xml'
            print "fname: " + fname           
            gtkimage.setFilename(fname)
            gtkimage.getWindow().set_title(os.path.basename(fname))
        dialog.destroy()
    if fname is not None:
        try:
            imageio.save_image(gtkimage)
        except (IOError, OSError), e:
            print "Error writing file: " + str(e)
        except StandardError, e:
            print "Error: " + str(e)
    return gtk.FALSE

def file_save_as_cb(menuitem, gtkimage):
    save_flag = False
    dialog = gtk.FileSelection("Save As ...")
    window = gtkimage.getWindow()
    dialog.set_transient_for(window)
    fname = gtkimage.getFilename()
    if fname is None:
        fname = window.get_title()
    dialog.set_filename(fname)
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        fname = dialog.get_filename()
        if fname == "":
            fname = 'Untitled.xml'
        if not fname.endswith('.xml'):
            fname = fname + '.xml'
        gtkimage.setFilename(fname)
        gtkimage.getWindow().set_title(os.path.basename(fname))
        save_flag = True
    dialog.destroy()
    if save_flag:
        #
        # there should be some effort made to avoid overwriting
        # an existing file ...
        #
        try:
            imageio.save_image(gtkimage)
        except (IOError, OSError), e:
            print"Error writing file: " + str(e)
        except StandardError, e:
            print "Error: " + str(e)
    return gtk.FALSE

def file_save_layer_cb(menuitem, gtkimage):
    # print "called file_save_layer_cb()"
    active = gtkimage.getActiveLayer()
    layer_name = active.getName()
    dialog = gtk.FileSelection("Save Layer As ...")
    dialog.set_transient_for(gtkimage.getWindow())
    dialog.set_filename(layer_name)
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        fname = dialog.get_filename()
        print "Saving layer as '%s'" % fname
        #
        # fixme - add the layer saving code ...
        #
    dialog.destroy()
    return gtk.FALSE

def file_quit_cb(menuitem, gtkimage):
    gtk.mainquit()
    return gtk.FALSE

def edit_menu_init_cb(menuitem, gtkimage):
    if gtkimage.hasSelection():
        submenu = menuitem.get_submenu()
    return gtk.FALSE

def edit_copy_cb(menuitem, gtkimage):
    if gtkimage.hasSelection():
        for lyr, obj in gtkimage.getSelectedObjects():
            globals.selectobj.storeObject(gtkimage, lyr, obj)
    return gtk.FALSE

def edit_cut_cb(menuitem, gtkimage):
    if gtkimage.hasSelection():
        for lyr, obj in gtkimage.getSelectedObjects():
            globals.selectobj.storeObject(gtkimage, lyr, obj)
            #
            # check that the object is still in the layer - objects like
            # dimensions get deleted if the points they refer to are
            # deleted ...
            #
            if obj in lyr:
                gtkimage.delObject(obj,lyr)
        gtkimage.redraw()
    return gtk.FALSE

def edit_paste_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click where you want to paste the objects")
    gtkedit.paste_mode_init(gtkimage.getTool())
    return gtk.FALSE

def select_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the items you want to select.")
    gtkedit.select_mode_init(gtkimage.getTool())
    return gtk.FALSE

def select_all_points_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _point in _active_layer.getLayerEntities("point"):
        gtkimage.selectObject(_active_layer, _point)
    return gtk.FALSE

def select_all_segments_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _segment in _active_layer.getLayerEntities("segment"):
        gtkimage.selectObject(_active_layer, _segment)
    return gtk.FALSE

def select_all_circles_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _circle in _active_layer.getLayerEntities("circle"):
        gtkimage.selectObject(_active_layer, _circle)
    return gtk.FALSE

def select_all_arcs_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _arc in _active_layer.getLayerEntities("arc"):
        gtkimage.selectObject(_active_layer, _arc)
    return gtk.FALSE

def select_all_hclines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _hcl in _active_layer.getLayerEntities("hcline"):
        gtkimage.selectObject(_active_layer, _hcl)
    return gtk.FALSE

def select_all_vclines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _vcl in _active_layer.getLayerEntities("vcline"):
        gtkimage.selectObject(_active_layer, _vcl)
    return gtk.FALSE

def select_all_aclines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _acl in _active_layer.getLayerEntities("acline"):
        gtkimage.selectObject(_active_layer, _acl)
    return gtk.FALSE

def select_all_clines_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _cline in _active_layer.getLayerEntities("cline"):
        gtkimage.selectObject(_active_layer, _cline)
    return gtk.FALSE

def select_all_ccircles_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _ccircle in _active_layer.getLayerEntities("ccircle"):
        gtkimage.selectObject(_active_layer, _ccircle)
    return gtk.FALSE

def select_all_leaders_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _leader in _active_layer.getLayerEntities("leader"):
        gtkimage.selectObject(_active_layer, _leader)
    return gtk.FALSE

def select_all_ldims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _ldim in _active_layer.getLayerEntities("linear_dimension"):
        gtkimage.selectObject(_active_layer, _ldim)
    return gtk.FALSE

def select_all_hdims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _hdim in _active_layer.getLayerEntities("horizontal_dimension"):
        gtkimage.selectObject(_active_layer, _hdim)
    return gtk.FALSE

def select_all_vdims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _vdim in _active_layer.getLayerEntities("vertical_dimension"):
        gtkimage.selectObject(_active_layer, _vdim)
    return gtk.FALSE

def select_all_rdims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _rdim in _active_layer.getLayerEntities("radial_dimension"):
        gtkimage.selectObject(_active_layer, _rdim)
    return gtk.FALSE

def select_all_adims_cb(menuitem, gtkimage):
    _active_layer = gtkimage.getActiveLayer()
    for _adim in _active_layer.getLayerEntities("angular_dimension"):
        gtkimage.selectObject(_active_layer, _adim)
    return gtk.FALSE

def prefs_cb(menuitem, gtkimage):
    gtkprefs.check_prefs(gtkimage)
    return gtk.FALSE

def draw_menu_init_cb(menuitem, gtkimage):
    # print "in draw_menu_init_cb()"
    return gtk.FALSE

def draw_point_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter the point.")
    gtkentities.point_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_segment_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a point.")
    gtkentities.segment_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_rectangle_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a point.")
    gtkentities.rectangle_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_circle_center_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a point.")
    gtkentities.circle_center_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_circle_tp_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a point.")
    gtkentities.circle_tp_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_arc_center_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a point.")
    gtkentities.arc_center_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_hcl_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter y coordinate")
    gtkconobjs.hcline_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_vcl_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter x coordinate")
    gtkconobjs.vcline_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_acl_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    gtkconobjs.acline_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_cl_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    gtkconobjs.cline_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_perpendicular_cline(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the object you want a perpendicular to")
    gtkconobjs.perpendicular_cline_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_tangent_cline(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the circle object you want a tangent to")
    gtkconobjs.tangent_cline_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_ccirc_cp_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    gtkconobjs.ccircle_cpmode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_ccirc_tp_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter a Point")
    gtkconobjs.ccircle_tpmode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_chamfer_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the points where you want a chamfer.")    
    gtkentities.chamfer_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_fillet_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the points where you want a fillet.")
    gtkentities.fillet_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_leader_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area to enter the first point")
    gtkentities.leader_mode_init(gtkimage.getTool())
    return gtk.FALSE

def draw_set_style(menuitem, gtkimage):
    gtkentities.set_active_style(gtkimage)
    return gtk.FALSE

def draw_set_linetype(menuitem, gtkimage):
    gtkentities.set_active_linetype(gtkimage)
    return gtk.FALSE

def draw_set_color(menuitem, gtkimage):
    gtkentities.set_active_color(gtkimage)
    return gtk.FALSE

def draw_set_thickness(menuitem, gtkimage):
    gtkentities.set_line_thickness(gtkimage)
    return gtk.FALSE

def modify_menu_init_cb(menuitem, gtkimage):
    # print "in modify_menu_init_cb()"
    return gtk.FALSE

def move_horizontal_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter the distance.")
    gtkmodify.move_horizontal_init(gtkimage.getTool())
    return gtk.FALSE

def move_vertical_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter the distance.")
    gtkmodify.move_vertical_init(gtkimage.getTool())
    return gtk.FALSE

def move_twopoint_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area.")
    gtkmodify.move_twopoint_init(gtkimage.getTool())
    return gtk.FALSE

def stretch_horiz_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on a point or enter the horizontal distance.")
    gtkmodify.stretch_horizontal_init(gtkimage.getTool())
    return gtk.FALSE

def stretch_vert_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click in the drawing area or enter the distance.")
    gtkmodify.stretch_vertical_init(gtkimage.getTool())
    return gtk.FALSE

def stretch_twopoint_cb(menuitem, gtkimage):
    print "in stretch_twopoint_cb()" # fixme
    return gtk.FALSE

def transfer_object(menuitem, gtkimage):
    prompt = "Click on the objects you want to transfer to the active layer"
    gtkimage.setPrompt(prompt)
    gtkmodify.transfer_object_init(gtkimage.getTool())
    return gtk.FALSE

def split_object(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the objects you want to split")
    gtkmodify.split_object_init(gtkimage.getTool())
    return gtk.FALSE

def delete_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the items you want to delete.")
    gtkmodify.delete_mode_init(gtkimage.getTool())
    return gtk.FALSE

def change_style_cb(menuitem, gtkimage):
    gtkmodify.change_style_init(gtkimage)
    return gtk.FALSE

def change_color_cb(menuitem, gtkimage):
    gtkmodify.change_color_init(gtkimage)
    return gtk.FALSE

def change_linetype_cb(menuitem, gtkimage):
    gtkmodify.change_linetype_init(gtkimage)
    return gtk.FALSE

def change_thickness_cb(menuitem, gtkimage):
    gtkmodify.change_thickness_init(gtkimage)
    return gtk.FALSE

def zoom_cb(menuitem, gtkimage):
    # print "in zoom_cb()"
    gtkimage.setPrompt("Click in the window.")
    gtkmodify.zoom_init(gtkimage.getTool())
    return gtk.FALSE

def zoom_in_cb(menuitem, gtkimage):
    _xmin, _ymin, _xmax, _ymax = gtkimage.getView()
    _scale = gtkimage.getUnitsPerPixel()
    _xdiff = abs(_xmax - _xmin)
    _ydiff = abs(_ymax - _ymin)
    _xmin = (_xmin + _xmax)/2.0 - _xdiff/4.0
    _ymin = (_ymin + _ymax)/2.0 - _ydiff/4.0
    gtkimage.setView(_xmin, _ymin, (_scale/2.0))
    gtkimage.redraw()
    return gtk.FALSE

def zoom_out_cb(menuitem, gtkimage):
    _xmin, _ymin, _xmax, _ymax = gtkimage.getView()
    _scale = gtkimage.getUnitsPerPixel()
    _xdiff = abs(_xmax - _xmin)
    _ydiff = abs(_ymax - _ymin)
    _xmin = (_xmin + _xmax)/2.0 - _xdiff
    _ymin = (_ymin + _ymax)/2.0 - _ydiff
    gtkimage.setView(_xmin, _ymin, (_scale * 2.0))
    gtkimage.redraw()
    return gtk.FALSE

def zoom_fit_cb(menuitem, gtkimage):
    gtkimage.fitImage()
    return gtk.FALSE

def dimension_linear_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkdimension.linear_mode_init(gtkimage.getTool())
    return gtk.FALSE

def dimension_menu_init_cb(menuitem, gtkimg):
    # print "in dimension_menu_init_cb()"
    return gtk.FALSE

def dimension_horizontal_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkdimension.horizontal_mode_init(gtkimage.getTool())
    return gtk.FALSE

def dimension_vertical_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on the first point for the dimension.")
    gtkdimension.vertical_mode_init(gtkimage.getTool())
    return gtk.FALSE

def dimension_radial_cb(menuitem, gtkimage):
    gtkimage.setPrompt("Click on an arc or a circle to dimension.")
    gtkdimension.radial_mode_init(gtkimage.getTool())
    return gtk.FALSE

def dimension_angular_cb(menuitem, gtkimage):
    # print "in dimension_angular_cb()"
    gtkimage.setPrompt("Click on the angle vertex or an arc.")
    gtkdimension.angular_mode_init(gtkimage.getTool())
    return gtk.FALSE

def dimension_prefs_cb(menuitem, gtkimage):
    # print "in dimension_prefs_cb()"
    gtkdimprefs.check_dim_prefs(gtkimage)
    return gtk.FALSE

def cb(menuitem, window):
    print "called cb"
    return gtk.FALSE

#
# dictionary to bind callbacks
#

callbacks = {
    'file_menu_init' : file_menu_init_cb,
    'file_new' : file_new_cb,
    'file_open' : file_open_cb,
    'file_close' : file_close_cb,
    'file_save' : file_save_cb,
    'file_save_as' : file_save_as_cb,
    'file_save_layer' : file_save_layer_cb,
    'file_quit' : file_quit_cb,
    'edit_menu_init' : edit_menu_init_cb,
    'edit_cut' : edit_cut_cb,
    'edit_copy' : edit_copy_cb,
    'edit_paste' : edit_paste_cb,
    'select_cb' : select_cb,
    'select_all_points' : select_all_points_cb,
    'select_all_segments' : select_all_segments_cb,
    'select_all_circles' : select_all_circles_cb,
    'select_all_arcs' : select_all_arcs_cb,
    'select_all_hclines' : select_all_hclines_cb,
    'select_all_vclines' : select_all_vclines_cb,
    'select_all_aclines' : select_all_aclines_cb,
    'select_all_clines' : select_all_clines_cb,
    'select_all_ccircles' : select_all_ccircles_cb,
    'select_all_ldims' : select_all_ldims_cb,
    'select_all_hdims' : select_all_hdims_cb,
    'select_all_vdims' : select_all_vdims_cb,
    'select_all_rdims' : select_all_rdims_cb,
    'select_all_adims' : select_all_adims_cb,
    'prefs' : prefs_cb,
    'draw_menu_init' : draw_menu_init_cb,
    'draw_point' : draw_point_cb,
    'draw_segment' : draw_segment_cb,
    'draw_rectangle' : draw_rectangle_cb,
    'draw_circle_center' : draw_circle_center_cb,
    'draw_circle_tp' : draw_circle_tp_cb,
    'draw_arc_center' : draw_arc_center_cb,
    'draw_hcl' : draw_hcl_cb,
    'draw_vcl' : draw_vcl_cb,
    'draw_acl' : draw_acl_cb,
    'draw_cl' : draw_cl_cb,
    'draw_perpendicular_cline' : draw_perpendicular_cline,    
    'draw_tangent_cline' : draw_tangent_cline,
    'draw_ccirc_cp' : draw_ccirc_cp_cb,
    'draw_ccirc_tp' : draw_ccirc_tp_cb,
    'draw_chamfer' : draw_chamfer_cb,
    'draw_fillet' : draw_fillet_cb,
    'draw_leader' : draw_leader_cb,
    'draw_set_style' : draw_set_style,
    'draw_set_linetype' : draw_set_linetype,
    'draw_set_color': draw_set_color,
    'draw_set_thickness' : draw_set_thickness,
    'modify_menu_init' : modify_menu_init_cb,
    'transfer' : transfer_object,
    'split' : split_object,
    'delete' : delete_cb,
    'zoom' : zoom_cb,
    'zoom_in' : zoom_in_cb,
    'zoom_out' : zoom_out_cb,
    'zoom_fit' : zoom_fit_cb,
    'move_horiz' : move_horizontal_cb,
    'move_vert' : move_vertical_cb,
    'move_twopoint' : move_twopoint_cb,
    'stretch_horiz' : stretch_horiz_cb,
    'stretch_vert' : stretch_vert_cb,
    'stretch_twopoint' : stretch_twopoint_cb,
    'change_style' : change_style_cb,
    'change_color' : change_color_cb,
    'change_linetype' : change_linetype_cb,
    'change_thickness' : change_thickness_cb,
    'dimension_menu_init' : dimension_menu_init_cb,
    'dimension_linear' : dimension_linear_cb,
    'dimension_horiz' : dimension_horizontal_cb,
    'dimension_vert' : dimension_vertical_cb,
    'dimension_rad' : dimension_radial_cb,
    'dimension_ang' : dimension_angular_cb,
    'dimension_prefs' : dimension_prefs_cb,
    'cb' : cb
    }

#
# create a Menu object for a list-based menu representation
#

def build_menu(menulist, gtkimage):
    if not isinstance(menulist, list):
        raise TypeError, "Invalid list for build_menu(): " + `menulist`
    menu = gtk.Menu()
    for i in xrange(len(menulist)):
        name = menulist[i][0]
        if name is None:
            item = gtk.SeparatorMenuItem()
        else:
            obj = menulist[i][1]
            if Generic.menus.submenus.has_key(obj):
                cbkey = menulist[i][2]
                mlist = Generic.menus.submenus[obj]
                item = gtk.MenuItem(name)
                sub = build_menu(mlist, gtkimage)
                item.set_submenu(sub)
            else:
                if not callbacks.has_key(obj):
                    raise KeyError, "Missing callback: " + `obj`
                icon = menulist[i][2]
                accel = menulist[i][3]
                if icon is None:
                    item = gtk.MenuItem(name)
                else:
                    item = gtk.ImageMenuItem(name)
                    if iconmap.has_key(icon):
                        image = gtk.image_new_from_stock(iconmap[icon],
                                                         gtk.ICON_SIZE_MENU)
                        item.set_image(image)
                if accel is not None and accelmap.has_key(accel):
                    item.add_accelerator("activate", gtkimage.accel,
                                         accelmap[accel],
                                         gtk.gdk.CONTROL_MASK,
                                         gtk.ACCEL_VISIBLE)
                #
                # turn off menu items that don't do anything - this
                # is a temporary means of presenting the menus with
                # the items in their positions but make them unselectable
                #
                # as the code to handle them is added, remove them
                # from the test condition ...
                #
                if (obj == 'cb' or obj == 'file_save_layer'):
                    item.set_sensitive(gtk.FALSE)
                else:
                    item.connect("activate", callbacks[obj] , gtkimage)
        menu.append(item)
    return menu

#
# fill a menubar with the menu template info
#

def init_menu(menuitem, gtkimg):
    # print "in init_menu()"
    return gtk.FALSE

def fill_menubar(mb, gtkimage):
    if not isinstance(mb, gtk.MenuBar):
        raise TypeError, "Invalid gtk.MenuBar object: " + `mb`
    for i in xrange(len(Generic.menus.primary_menus)):
        name, key, cbkey = Generic.menus.primary_menus[i]
        assert Generic.menus.submenus.has_key(key), "Missing primary menu key!"
        if not callbacks.has_key(cbkey):
            raise ValueError, "Missing menu initializing key: " + `cbkey`
        mlist = Generic.menus.submenus[key]
        item = gtk.MenuItem(name)
        item.connect("activate", callbacks[cbkey], gtkimage)
        sub = build_menu(mlist, gtkimage)
        item.set_submenu(sub)
        mb.append(item)
