#
# Copyright (c) 2002, 2003 Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# functions for doing modifications on drawing
# entities
#

import math

import gtk

import Generic.dimension as dimension
import Generic.color as color
import Generic.graphicobject as graphicobject
import Generic.point as point
import Generic.segment as segment
import Generic.circle as circle
import Generic.ccircle as ccircle
import Generic.arc as arc
import Generic.intersections as intersections

#
# common code
#

def select_motion_notify(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    # _width, _height = gtkimage.getSize()
    _gc = gtkimage.getGC()
    _x = event.x
    _y = event.y
    _cp = tool.getCurrentPoint()
    if _cp is not None:
        _xc, _yc = _cp
        _xmin = min(_xc, _px)
        _ymin = min(_yc, _py)
        _rw = abs(_xc - _px)
        _rh = abs(_yc - _py)
        widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)
    tool.setCurrentPoint(_x, _y)
    _xmin = min(_x, _px)
    _ymin = min(_y, _py)
    _rw = abs(_x - _px)
    _rh = abs(_y - _py)
    widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)

def entry_get_distance(text):
    try:
        _text_dist = eval(text)
    except:
        raise ValueError, "Invalid text: " + `text`
    _dist = _text_dist
    if not isinstance(_dist, float):
        try:
            _dist = float(_text_dist)
        except:
            raise ValueError, "Invalid distance: " + `_text_dist`
    return _dist
            
#
# move objects
#

def move_button_press(gtkimage, tool):
    _x, _y = gtkimage.getPoint()
    #
    # need to find if the point is an intersection of drawing objects ...
    #
    tool.pushObject(_x)
    tool.pushObject(_y)

def move_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
    if len(_objs):
        _dx, _dy = tool.getShift()
        _active_layer.moveObjects(_objs, _dx, _dy)
        _active_layer.update()
        gtkimage.redraw()
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", move_elem_button_press_cb)
    
def move_elem_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _objdict = gtkimage.mapPoint(_x, _y, _tol)
    if len(_objdict):
        _active_layer = gtkimage.getActiveLayer()
        if _active_layer in _objdict:
            _objs = []
            for _obj, _pt in _objdict[_active_layer]:
                _objs.append(_obj)
            _dx, _dy = tool.getShift()
            _active_layer.moveObjects(_objs, _dx, _dy)
            _active_layer.update()
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", move_end_button_press_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)

def move_horiz_entry_event(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    try:
        dist = entry_get_distance(text)
    except:
        raise StandardError, "Invalid distance: " + `text`
    tool.setShift(dist, 0.0)
    gtkimage.setPrompt("Click on the objects to move.")
    tool.delHandler("entry_event")
    tool.setHandler("button_press", move_elem_button_press_cb)

#
# move horizontal
#

def move_horiz_button_press_cb(gtkimage, widget, event, tool):
    print "called move_horiz_button_press_cb()"
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the horizontal distance")
    else:
        x, y = gtkimage.getPoint()
        #
        # need to find if the point is an intersection of drawing objects ...
        #
        y1 = tool.popObject()
        x1 = tool.popObject()
        tool.setShift((x - x1), 0.0)
        gtkimage.setPrompt("Click on the objects to move.")
        tool.delHandler("entry_event")
        tool.setHandler("button_press", move_elem_button_press_cb)

def move_horizontal_init(tool):
    tool.initialize()
    tool.setHandler("button_press", move_horiz_button_press_cb)
    tool.setHandler("entry_event", move_horiz_entry_event)

#
# move vertical
#

def move_vert_entry_event(gtkimage, widget, tool):
    entry = gtkimage.getEntry()
    text = entry.get_chars(0,-1)
    entry.delete_text(0,-1)
    try:
        dist = entry_get_distance(text)
    except:
        raise StandardError, "Invalid distance: " + `text`
    tool.setShift(0.0, dist)
    gtkimage.setPrompt("Click on the objects to move.")
    tool.delHandler("entry_event")
    tool.setHandler("button_press", move_elem_button_press_cb)
    
def move_vert_button_press_cb(gtkimage, widget, event, tool):
    print "called move_vert_button_press_cb()"
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the vertical distance")
    else:
        x, y = gtkimage.getPoint()
        #
        # need to find if the point is an intersection of drawing objects ...
        #
        y1 = tool.popObject()
        x1 = tool.popObject()
        tool.setShift(0.0, (y - y1))
        gtkimage.setPrompt("Click on the objects to move.")
        tool.delHandler("entry_event")
        tool.setHandler("button_press", move_elem_button_press_cb)

def move_vertical_init(tool):
    tool.initialize()
    tool.setHandler("button_press", move_vert_button_press_cb)
    tool.setHandler("entry_event", move_vert_entry_event)

#
# move based on two mouse clicks
#

def move_tp_button_press_cb(gtkimage, widget, event, tool):
    print "called move_tp_button_press_cb()"
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the distance")
    else:
        x, y = gtkimage.getPoint()
        #
        # need to find if the point is an intersection of drawing objects ...
        #
        y1 = tool.popObject()
        x1 = tool.popObject()
        tool.setShift((x - x1), (y - y1))
        gtkimage.setPrompt("Click on the objects to move.")
        tool.setHandler("button_press", move_elem_button_press_cb)

def move_twopoint_init(tool):
    tool.initialize()
    tool.setHandler("button_press", move_tp_button_press_cb)

#
# delete objects
#

def delete_region_end_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    tool.setHandler("button_press", delete_button_press_cb)
    _active_layer = gtkimage.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
    gtkimage.refresh()
    if len(_objs):
        _active_layer.delObjects(_objs)
        gtkimage.redraw()
    delete_mode_init(tool)
    
def delete_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _active_layer = gtkimage.getActiveLayer()
    _tol = gtkimage.getTolerance()
    _objs = _active_layer.mapPoint((_x, _y), _tol)
    if len(_objs):
        _dims = []
        for _obj in _objs:
            if isinstance(_obj, dimension.Dimension):
                _dims.append(_obj)
            elif isinstance(_obj, tuple):
                _entity, _pt = _obj
                _active_layer.delObject(_entity)
            else:
                raise TypeError, "Unhandled object: " + `_obj`
        for _dim in _dims:
            if _dim in _active_layer: # it may have been removed ...
                _active_layer.delObject(_dim)
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.storeCoords(_x, _y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", delete_region_end_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)
        
def delete_mode_init(tool):
    tool.initialize()
    tool.setHandler("button_press", delete_button_press_cb)

#
# stretch operations
#

def stretch_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _dx, _dy = tool.getShift()
    for _pt in _active_layer.getLayerEntities("point"):
        if _pt.x > _xmax:
            break
        if _pt.inRegion(_xmin, _ymin, _xmax, _ymax):
            _pt.move(_dx, _dy)
    gtkimage.redraw()
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", stretch_elem_button_press_cb)

def stretch_elem_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _pt = _active_layer.hasPoint((_x, _y), _tol)
    if _pt is not None:
        _dx, _dy = tool.getShift()
        _pt.move(_dx, _dy)
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", stretch_end_button_press_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)

#
# stretch horizontal
#

def stretch_horiz_button_press_cb(gtkimage, widget, event, tool):
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the horizontal distance")
    else:
        _x, _y = gtkimage.getPoint()
        #
        # see if the point is at an intersection of drawing objects ...
        #
        _y1 = tool.popObject()
        _x1 = tool.popObject()
        tool.setShift((_x - _x1), 0.0)
        gtkimage.setPrompt("Select the points to move.")
        # tool.delHandler("entry_event")
        tool.setHandler("button_press", stretch_elem_button_press_cb)

def stretch_horizontal_init(tool):
    tool.initialize()
    tool.setHandler("button_press", stretch_horiz_button_press_cb)
    # tool.setHandler("entry_event", stretch_horiz_entry_event)

#
# stretch vertical
#

def stretch_vert_button_press_cb(gtkimage, widget, event, tool):
    if not len(tool):
        move_button_press(gtkimage, tool)
        gtkimage.setPrompt("Click a second point to indicate the vertical distance")
    else:
        _x, _y = gtkimage.getPoint()
        #
        # see if the point is at an intersection of drawing objects ...
        #
        _y1 = tool.popObject()
        _x1 = tool.popObject()
        tool.setShift(0.0, (_y - _y1))
        gtkimage.setPrompt("Select the points to move.")
        # tool.delHandler("entry_event")
        tool.setHandler("button_press", stretch_elem_button_press_cb)

def stretch_vertical_init(tool):
    tool.initialize()
    tool.setHandler("button_press", stretch_vert_button_press_cb)
    # tool.setHandler("entry_event", stretch_horiz_entry_event)

#
# split objects into two pieces or at intersection points
#

def _split_seg(seg, pt):
    _p1, _p2 = seg.getEndpoints()
    _style = seg.getStyle()
    _linetype = seg.getLinetype()
    _color = seg.getColor()
    _thickness = seg.getThickness()
    _s1 = segment.Segment(_p1, pt, _style, _linetype, _color, _thickness)
    _s2 = segment.Segment(pt, _p2, _style, _linetype, _color, _thickness)
    return (_s1, _s2)

def _split_seg_2pts(seg, pt1, pt2):
    _p1, _p2 = seg.getEndpoints()
    _style = seg.getStyle()
    _linetype = seg.getLinetype()
    _color = seg.getColor()
    _thickness = seg.getThickness()
    _s1 = segment.Segment(_p1, pt1, _style, _linetype, _color, _thickness)
    _s2 = segment.Segment(pt1, pt2, _style, _linetype, _color, _thickness)
    _s3 = segment.Segment(pt2, _p2, _style, _linetype, _color, _thickness)
    return (_s1, _s2, _s3)

def _split_circ(circ, pt):
    _cp = circ.getCenter()
    _rad = circ.getRadius()
    _cx, _cy = _cp.getCoords()
    _px, _py = pt.getCoords()
    _angle = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle < 0.0:
        _angle = _angle + 360.0
    _arc = arc.Arc(_cp, _rad, _angle, _angle,
                   circ.getStyle(), circ.getLinetype(),
                   circ.getColor(), circ.getThickness())
    return _arc

def _split_circ_2pts(circ, p1, p2):
    _cp = circ.getCenter()
    _r = circ.getRadius()
    _cx, _cy = _cp.getCoords()            
    _px, _py = p1.getCoords()
    _angle1 = (180.0/math.pi) * math.atan2((_py - _cy),(_px - _cx))
    if _angle1 < 0.0:
        _angle1 = _angle1 + 360.0
    _px, _py = p2.getCoords()
    _angle2 = (180.0/math.pi) * math.atan2((_py - _cy),(_px - _cx))
    if _angle2 < 0.0:
        _angle2 = _angle2 + 360.0
    _style = circ.getStyle()
    _color = circ.getColor()
    _linetype = circ.getLinetype()
    _thickness = circ.getThickness()
    _a1 = arc.Arc(_cp, _r, _angle1, _angle2, _style,
                  _linetype, _color, _thickness)
    _a2 = arc.Arc(_cp, _r, _angle2, _angle1, _style,
                  _linetype, _color, _thickness)
    return (_a1, _a2)

def _split_arc(a, pt):
    _cp = a.getCenter()
    _rad = a.getRadius()
    _sa = a.getStartAngle()
    _ea = a.getEndAngle()
    _cx, _cy = _cp.getCoords()
    _px, _py = pt.getCoords()
    _angle = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle < 0.0:
        _angle = _angle + 360.0
    _style = a.getStyle()
    _linetype = a.getLinetype()
    _color = a.getColor()
    _thickness = a.getThickness()
    _arc1 = arc.Arc(_cp, _rad, _sa, _angle,
                    _style, _linetype, _color, _thickness)
    _arc2 = arc.Arc(_cp, _rad, _angle, _ea,
                    _style, _linetype, _color, _thickness)
    return (_arc1, _arc2)

def _split_arc_2pts(a, pt1, pt2):
    _cp = a.getCenter()
    _rad = a.getRadius()
    _sa = a.getStartAngle()
    _ea = a.getEndAngle()
    _cx, _cy = _cp.getCoords()
    _px, _py = pt1.getCoords()
    _angle1 = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle1 < 0.0:
        _angle1 = _angle1 + 360.0
    _px, _py = pt2.getCoords()
    _angle2 = (180.0/math.pi) * math.atan2((_py - _cy), (_px - _cx))
    if _angle2 < 0.0:
        _angle2 = _angle2 + 360.0
    if _angle1 < _angle2:
        _a1 = _angle1
        _a2 = _angle2
    else:
        _a1 = _angle2
        _a2 = _angle1
    _style = a.getStyle()
    _linetype = a.getLinetype()
    _color = a.getColor()
    _thickness = a.getThickness()
    _arc1 = arc.Arc(_cp, _rad, _sa, _a1,
                    _style, _linetype, _color, _thickness)
    _arc2 = arc.Arc(_cp, _rad, _a1, _a2,
                    _style, _linetype, _color, _thickness)
    _arc3 = arc.Arc(_cp, _rad, _a2, _ea,
                    _style, _linetype, _color, _thickness)    
    return (_arc1, _arc2, _arc3)

def _pt_in_new_points(newpts, pt):
    _newpt = None
    for _pt in newpts:
        if pt == _pt:
            _newpt = _pt
            break
    return _newpt
    
def _seg_seg_split(newpts, newobjs, lyr, seg1, seg2):
    assert isinstance(seg1, segment.Segment), "seg1 wrong type: " + `seg1`
    assert isinstance(seg2, segment.Segment), "seg2 wrong type: " + `seg2`
    _hit_seg1 = False
    _hit_seg2 = False
    _ipts = intersections.findIntersections(seg1, seg2)
    if len(_ipts):
        assert len(_ipts) == 1, "Invalid seg/seg int length: %d" + len(_ipts)
        _ip = _ipts[0]
        _p1, _p2 = seg1.getEndpoints()
        _p3, _p4 = seg2.getEndpoints()
        _lp = lyr.hasPoint(_ip)
        if _lp is None:
            _lp = _pt_in_new_points(newpts, _ip)
            if _lp is None:
                _lp = _ip
                newpts.append(_lp)
        if _lp != _p1 and _lp != _p2:
            _hit_seg1 = True
            _s1, _s2 = _split_seg(seg1, _lp)
            newobjs[_s1] = True
            newobjs[_s2] = True
        if _lp != _p3 and _lp != _p4:
            _hit_seg2 = True
            _s1, _s2 = _split_seg(seg2, _lp)
            newobjs[_s1] = True
            newobjs[_s2] = True
    return (_hit_seg1, _hit_seg2)

def _seg_circ_split(newpts, newobjs, lyr, seg, circ):
    assert isinstance(seg, segment.Segment), "seg wrong type: " + `seg`
    assert isinstance(circ, circle.Circle), "circ wrong type: " + `circ`
    assert not isinstance(circ, arc.Arc), "arc in _seg_circ_split()"
    _hit_seg = False
    _hit_circ = False
    _ipts = intersections.findIntersections(seg, circ)
    if len(_ipts):
        _count = len(_ipts)
        _p1, _p2 = seg.getEndpoints()
        if _count == 1:
            _hit_circ = True
            _ip = _ipts[0]
            _arc = _split_circ(circ, _ip)
            newobjs[_arc] = True
            if _ip != _p1 and _ip != _p2:
                _hit_seg = True
                _lp = lyr.hasPoint(_ip)
                if _lp is None:
                    _lp = _pt_in_new_points(newpts, _ip)
                    if _lp is None:
                        _lp = _ip
                        newpts.append(_lp)
                _s1, _s2 = _split_seg(seg, _lp)
                newobjs[_s1] = True
                newobjs[_s2] = True
        elif _count == 2:
            _hit_circ = True
            _arc1, _arc2 = _split_circ_2pts(circ, _ipts[0], _ipts[1])
            newobjs[_arc1] = True
            newobjs[_arc2] = True
            _ip1 = _ipts[0]
            _ip2 = _ipts[1]
            if (_ip1 != _p1 and
                _ip1 != _p2 and
                _ip2 != _p1 and
                _ip2 != _p2):
                _hit_seg = True
                if ((_p1 - _ip1) < (_p1 - _ip2)):
                    _m1 = _ip1
                    _m2 = _ip2
                else:
                    _m1 = _ip2
                    _m2 = _ip1
                _lp1 = lyr.hasPoint(_m1)
                if _lp1 is None:
                    _lp1 = _pt_in_new_points(newpts, _m1)
                    if _lp1 is None:
                        _lp1 = _m1
                        newpts.append(_lp1)
                _lp2 = lyr.hasPoint(_m2)
                if _lp2 is None:
                    _lp2 = _pt_in_new_points(newpts, _m2)
                    if _lp2 is None:
                        _lp2 = _m2
                        newpts.append(_lp2)
                _s1, _s2, _s3 = _split_seg_2pts(seg, _lp1, _lp2)
                newobjs[_s1] = True
                newobjs[_s2] = True
                newobjs[_s3] = True
            elif _ip1 == _p1 or _ip1 == _p2:
                if _ip2 != _p1 and _ip2 != _p2:
                    _hit_seg = True
                    _lp = lyr.hasPoint(_ip2)
                    if _lp is None:
                        _lp = _pt_in_new_points(newpts, _ip2)
                        if _lp is None:
                            _lp = _ip2
                            newpts.append(_lp)
                    _s1, _s2 = _split_seg(seg, _lp)
                    newobjs[_s1] = True
                    newobjs[_s2] = True
            elif _ip2 == _p1 or _ip2 == _p2:
                if _ip1 != _p1 and _ip1 != _p2:
                    _hit_seg = True
                    _lp = lyr.hasPoint(_ip1)
                    if _lp is None:
                        _lp = _pt_in_new_points(newpts, _ip1)
                        if _lp is None:
                            _lp = _ip1
                            newpts.append(_lp)
                    _s1, _s2 = _split_seg(seg, _lp)
                    newobjs[_s1] = True
                    newobjs[_s2] = True
        else:
            raise ValueError, "Unexpected seg/circle int count: %d" % _count
    return (_hit_seg, _hit_circ)

def _seg_arc_split(newpts, newobjs, lyr, seg, a):
    assert isinstance(seg, segment.Segment), "seg wrong type: " + `seg`
    assert isinstance(a, arc.Arc), "arc wrong type: " + `a`
    _hit_seg = False
    _hit_arc = False
    _ipts = intersections.findIntersections(seg, a)
    if len(_ipts):
        _count = len(_ipts)
        _p1, _p2 = seg.getEndpoints()
        _cp = a.getCenter()
        _ep1, _ep2 = a.getEndpoints()
        _cx, _cy = _cp.getCoords()
        if _count == 1:
            _ip = _ipts[0]
            if _ip != _p1 and _ip != _p2:
                _hit_seg = True
                _lp = lyr.hasPoint(_ip)
                if _lp is None:
                    _lp = _pt_in_new_points(newpts, _ip)
                    if _lp is None:
                        _lp = _ip
                        newpts.append(_lp)
                _s1, _s2 = _split_seg(seg, _lp)
                newobjs[_s1] = True
                newobjs[_s2] = True
            if _ip != _ep1 and _ip != _ep2:
                _hit_arc = True
                _a1, _a2 = _split_arc(a, _ip)
                newobjs[_a1] = True
                newobjs[_a2] = True
        elif _count == 2:
            _ip1 = _ipts[0]
            _ip2 = _ipts[1]
            _angle1 = None
            _angle2 = None
            if _ip1 != _ep1 and _ip1 != _ep2:
                _ix, _iy = _ip1.getCoords()
                _angle1 = (180.0/math.pi) * math.atan2((_iy - _cy),(_ix - _cx))
                if _angle1 < 0.0:
                    _angle1 = _angle1 + 360.0
            if _ip2 != _ep1 and _ip2 != _ep2:
                _ix, _iy = _ip2.getCoords()
                _angle2 = (180.0/math.pi) * math.atan2((_iy - _cy),(_ix - _cx))
                if _angle2 < 0.0:
                    _angle2 = _angle2 + 360.0
            _angles = []
            if _angle1 is not None:
                _angles.append(_angle1)
            if _angle2 is not None:
                _angles.append(_angle2)
            if len(_angles):
                _hit_arc = True
                _sa = a.getStartAngle()
                _angles.append(_sa)
                _angles.append(a.getEndAngle())
                _angles.sort()
                while (abs(_angles[0] - _sa) > 1e-10):
                    _temp = _angles.pop()
                    _angles.insert(0, _temp)
                _r = a.getRadius()
                _style = a.getStyle()
                _color = a.getColor()
                _linetype = a.getLinetype()
                _thickness = a.getThickness()
                _a1 = arc.Arc(_cp, _r, _angles[0], _angles[1],
                              _style, _linetype, _color, _thickness)
                newobjs[_a1] = True
                _a2 = arc.Arc(_cp, _r, _angles[1], _angles[2],
                              _style, _linetype, _color, _thickness)
                newobjs[_a2] = True
                if len(_angles) == 4:
                    _a3 = arc.Arc(_cp, _r, _angles[2], _angles[3],
                                  _style, _linetype, _color, _thickness)
                    newobjs[_a3] = True
            if (_ip1 != _p1 and
                _ip1 != _p2 and
                _ip2 != _p1 and
                _ip2 != _p2):
                _hit_seg = True
                if ((_p1 - _ip1) < (_p1 - _ip2)):
                    _m1 = _ip1
                    _m2 = _ip2
                else:
                    _m1 = _ip2
                    _m2 = _ip1
                _lp1 = lyr.hasPoint(_m1)
                if _lp1 is None:
                    _lp1 = _pt_in_new_points(newpts, _m1)
                    if _lp1 is None:
                        _lp1 = _m1
                        newpts.append(_lp1)
                _lp2 = lyr.hasPoint(_m2)
                if _lp2 is None:
                    _lp2 = _pt_in_new_points(newpts, _m2)
                    if _lp2 is None:
                        _lp2 = _m2
                        newpts.append(_lp2)
                _style = seg.getStyle()
                _color = seg.getColor()
                _linetype = seg.getLinetype()
                _thickness = seg.getThickness()
                _s1 = segment.Segment(_p1, _lp1, _style,
                                      _linetype, _color, _thickness)
                _s2 = segment.Segment(_lp1, _lp2, _style,
                                      _linetype, _color, _thickness)
                _s3 = segment.Segment(_lp2, _p2, _style,
                                      _linetype, _color, _thickness)
                newobjs[_s1] = True
                newobjs[_s2] = True
                newobjs[_s3] = True
            elif _ip1 == _p1 or _ip1 == _p2:
                if _ip2 != _p1 and _ip2 != _p2:
                    _hit_seg = True
                    _lp = lyr.hasPoint(_ip2)
                    if _lp is None:
                        _lp = _pt_in_new_points(newpts, _ip2)
                        if _lp is None:
                            _lp = _ip2
                            newpts.append(_lp)
                    _s1, _s2 = _split_seg(seg, _lp)
                    newobjs[_s1] = True
                    newobjs[_s2] = True
            elif _ip2 == _p1 or _ip2 == _p2:
                if _ip1 != _p1 and _ip1 != _p2:
                    _hit_seg = True
                    _lp = lyr.hasPoint(_ip1)
                    if _lp is None:
                        _lp = _pt_in_new_points(newpts, _ip1)
                        if _lp is None:
                            _lp = _ip1
                            newpts.append(_lp)
                    _s1, _s2 = _split_seg(seg, _lp)
                    newobjs[_s1] = True
                    newobjs[_s2] = True
    return (_hit_seg, _hit_arc)

def _circ_circ_split(newpts, newobjs, lyr, circ1, circ2):
    assert isinstance(circ1, circle.Circle), "circ1 wrong type: " + `circ1`
    assert not isinstance(circ1, arc.Arc), "arc in _circ_circ_split()"
    assert isinstance(circ2, circle.Circle), "circ2 wrong type: " + `circ2`
    assert not isinstance(circ2, arc.Arc), "arc in _circ_circ_split()"
    _hit_circ1 = False
    _hit_circ2 = False
    _ipts = intersections.findIntersections(circ1, circ2)
    if len(_ipts):
        _hit_circ1 = True
        _hit_circ2 = True        
        _count = len(_ipts)
        if _count == 1:
            _ip = _ipts[0]
            _arc1 = _split_circ(circ1, _ip)
            newobjs[_arc1] = True
            _arc2 = _split_circ(circ2, _ip)
            newobjs[_arc2] = True
        elif _count == 2:
            _arc1, _arc2 = _split_circ_2pts(circ1, _ipts[0], _ipts[1])
            newobjs[_arc1] = True
            newobjs[_arc2] = True
            _arc3, _arc4 = _split_circ_2pts(circ2, _ipts[0], _ipts[1])
            newobjs[_arc3] = True
            newobjs[_arc4] = True            
        else:
            raise ValueError, "Unexpected circle/circle int count: %d" % _count
    return (_hit_circ1, _hit_circ2)

def _circ_arc_split(newpts, newobjs, lyr, circ, a):
    assert isinstance(circ, circle.Circle), "circ wrong type: " + `circ`
    assert not isinstance(circ, arc.Arc), "arc in _circ_circ_split()"
    assert isinstance(a, arc.Arc), "arc wrong type: " + `a`
    _hit_circ = False
    _hit_arc = False
    _ipts = intersections.findIntersections(circ, a)
    if len(_ipts):
        _hit_circ = True
        _count = len(_ipts)
        if _count == 1:
            _ip = _ipts[0]
            _arc1 = _split_circ(circ, _ip)
            newobjs[_arc1] = True
            _ep1, _ep2 = a.getEndpoints()
            if _ip != _ep1 and _ip != _ep2:
                _hit_arc = True
                _arc2, _arc3 = _split_arc(a, _ip)
                newobjs[_arc2] = True
                newobjs[_arc3] = True
        elif _count == 2:
            _arc1, _arc2 = _split_circ_2pts(circ, _ipts[0], _ipts[1])
            newobjs[_arc1] = True
            newobjs[_arc2] = True
            _arc3, _arc4, _arc5 = _split_arc_2pts(a, _ipts[0], _ipts[1])
            newobjs[_arc3] = True
            newobjs[_arc4] = True
            newobjs[_arc5] = True
        else:
            raise ValueError, "Unexpected circle/arc int count: %d" % _count
    return (_hit_circ, _hit_arc)

def _arc_arc_split(newpts, newobjs, lyr, arc1, arc2):
    assert isinstance(arc1, arc.Arc), "arc1 wrong type: " + `arc1`
    assert isinstance(arc2, arc.Arc), "arc2 wrong type: " + `arc2`
    _hit_arc1 = False
    _hit_arc2 = False
    _ipts = intersections.findIntersections(arc1, arc2)
    if len(_ipts):
        _ep1, _ep2 = arc1.getEndpoints()
        _ep3, _ep4 = arc2.getEndpoints()        
        _count = len(_ipts)
        if _count == 1:
            _ip = _ipts[0]
            if _ip != _ep1 and _ip != _ep2:
                _hit_arc1 = True
                _a1, _a2 = _split_arc(arc1, _ip)
                newobjs[_a1] = True
                newobjs[_a2] = True
            if _ip != _ep3 and _ip != _ep4:
                _hit_arc2 = True
                _a3, _a4 = _split_arc(arc2, _ip)
                newobjs[_a3] = True
                newobjs[_a4] = True
        elif _count == 2:
            _ip1 = _ipts[0]
            _ip2 = _ipts[1]
            if (_ip1 != _ep1 and
                _ip1 != _ep2 and
                _ip2 != _ep1 and
                _ip2 != _ep2):
                _hit_arc1 = True
                _a1, _a2, _a3 = _split_arc_2pts(arc1, _ip1, _ip2)
                newobjs[_a1] = True
                newobjs[_a2] = True
                newobjs[_a3] = True
            elif _ip1 == _ep1 or _ip1 == _ep2:
                if _ip2 != _ep1 and _ip2 != _ep2:
                    _hit_arc1 = True
                    _a1, _a2 = _split_arc(arc1, _ip2)
                    newobjs[_a1] = True
                    newobjs[_a2] = True
            elif _ip2 == _ep1 or _ip2 == _ep2:
                if _ip1 != _ep1 and _ip1 != _ep2:
                    _hit_arc1 = True
                    _a1, _a2 = _split_arc(arc1, _ip1)
                    newobjs[_a1] = True
                    newobjs[_a2] = True
            if (_ip1 != _ep3 and
                _ip1 != _ep4 and
                _ip2 != _ep3 and
                _ip2 != _ep4):
                _hit_arc2 = True
                _a4, _a5, _a6 = _split_arc_2pts(arc2, _ip1, _ip2)
                newobjs[_a4] = True
                newobjs[_a5] = True
                newobjs[_a6] = True
            elif _ip1 == _ep3 or _ip1 == _ep4:
                if _ip2 != _ep3 and _ip2 != _ep4:
                    _hit_arc2 = True
                    _a1, _a2 = _split_arc(arc2, _ip2)
                    newobjs[_a1] = True
                    newobjs[_a2] = True
            elif _ip2 == _ep3 or _ip2 == _ep4:
                if _ip1 != _ep3 and _ip1 != _ep4:
                    _hit_arc2 = True
                    _a1, _a2 = _split_arc(arc2, _ip1)
                    newobjs[_a1] = True
                    newobjs[_a2] = True
        else:
            raise ValueError, "Unexpected arc/arc int count: %d" % _count
    return (_hit_arc1, _hit_arc2)

def _split_objects(lyr, objlist):
    _newpts = []
    _newobjs = {}
    _delobjs = {}
    while len(objlist):
        _testobj = objlist.pop()
        _hit_test = False
        _hit_obj = False
        _idx = 0        
        for _obj in objlist:
            if isinstance(_testobj, segment.Segment):
                if isinstance(_obj, segment.Segment):
                    _hit_test, _hit_obj = _seg_seg_split(_newpts, _newobjs,
                                                         lyr, _testobj, _obj)
                elif isinstance(_obj, arc.Arc):
                    _hit_test, _hit_obj = _seg_arc_split(_newpts, _newobjs,
                                                         lyr, _testobj, _obj)
                elif isinstance(_obj, circle._Circle):
                    _hit_test, _hit_obj = _seg_circ_split(_newpts, _newobjs,
                                                          lyr, _testobj, _obj)
                else:
                    raise TypeError, "Seg/Obj unexpected: " + `_obj`
            elif isinstance(_testobj, arc.Arc):                
                if isinstance(_obj, segment.Segment):
                    _hit_obj, _hit_test = _seg_arc_split(_newpts, _newobjs,
                                                         lyr, _obj, _testobj)
                elif isinstance(_obj, arc.Arc):
                    _hit_test, _hit_obj = _arc_arc_split(_newpts, _newobjs,
                                                         lyr, _testobj, _obj)
                elif isinstance(_obj, circle._Circle):
                    _hit_obj, _hit_test = _circ_arc_split(_newpts, _newobjs,
                                                          lyr, _obj, _testobj)
                else:
                    raise TypeError, "Arc/Obj unexpected: " + `_obj`
            elif isinstance(_testobj, circle.Circle):
                if isinstance(_obj, segment.Segment):
                    _hit_obj, _hit_test = _seg_circ_split(_newpts, _newobjs,
                                                          lyr, _obj, _testobj)
                elif isinstance(_obj, arc.Arc):
                    _hit_test, _hit_obj = _circ_arc_split(_newpts, _newobjs,
                                                          lyr, _testobj, _obj)
                elif isinstance(_obj, circle._Circle):
                    _hit_test, _hit_obj = _circ_circ_split(_newpts, _newobjs,
                                                           lyr, _testobj, _obj)
                else:
                    raise TypeError, "Arc/Obj unexpected: " + `_obj`
            #
            # if we hit _obj or _testobj, remove them from the
            # _newobjs dict if they are there, or set them up
            # to be deleted from the layer
            #
            if _hit_obj:
                del objlist[_idx]
                if _obj in _newobjs:
                    del _newobjs[_obj]
                else:
                    _delobjs[_obj] = 1
            if _hit_test:
                if _testobj in _newobjs:
                    del _newobjs[_testobj]
                else:
                    _delobjs[_testobj] = 1
            #
            # add any new objects to the list of splittable objects
            #
            for _obj in _newobjs:
                if _newobjs[_obj]:
                    objlist.append(_obj)
                    _newobjs[_obj] = False
            if _hit_obj or _hit_test:
                break
            _idx = _idx + 1
    #
    # add all the new objects into the layer, and delete
    # the objects that have been split apart
    #
    for _pt in _newpts:
        lyr.addObject(_pt)
    for _obj in _newobjs:
        lyr.addObject(_obj)
    for _obj in _delobjs:
        lyr.delObject(_obj)

def split_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _objs = _active_layer.objsInRegion(_xmin, _ymin, _xmax, _ymax, True)
    gtkimage.refresh()
    if len(_objs):
        _splitable = []
        for _obj in _objs:
            if isinstance(_obj, segment.Segment):
                _splitable.append(_obj)
            elif isinstance(_obj, circle.Circle): # circles and arcs
                _splitable.append(_obj)
            else:
                pass
        if len(_splitable):
            _split_objects(_active_layer, _splitable)
        gtkimage.redraw()
    gtkimage.setPrompt("Click on the objects you want to split.")
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", split_object_button_press_cb)
    
def split_object_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _objlist = _active_layer.mapPoint((_x, _y), _tol, None)
    if len(_objlist):
        for _obj, _pt in _objlist:
            if isinstance(_obj, segment.Segment):
                _p1, _p2 = _obj.getEndpoints()
                _lpt = _active_layer.hasPoint(_pt)
                if _lpt is None:
                    _active_layer.addObject(_pt)
                    _lpt = _pt
                _s1, _s2 = _split_seg(_obj, _pt)
                _active_layer.addObject(_s1)
                _active_layer.addObject(_s2)
                _active_layer.delObject(_obj)
            elif isinstance(_obj, arc.Arc):
                _arc1, _arc2 = _split_arc(_obj, _pt)
                _active_layer.addObject(_arc1)
                _active_layer.addObject(_arc2)
                _active_layer.delObject(_obj)
            elif isinstance(_obj, circle.Circle):
                _active_layer.addObject(_split_circ(_obj, _pt))
                _active_layer.delObject(_obj)
            else:
                print "obj: " + `_obj`
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", split_end_button_press_cb)
        _gc = gtkimage.getGC()
        _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                                gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        _gc.set_function(gtk.gdk.INVERT)
        
def split_object_init(tool):
    tool.initialize()
    tool.setHandler("button_press", split_object_button_press_cb)

#
# transfer objects from one layer to another
#

def transfer_end_button_press_cb(gtkimage, widget, event, tool):
    _x2, _y2 = gtkimage.getPoint()
    _y1 = tool.popObject()
    _x1 = tool.popObject()
    _xmin = min(_x1, _x2)
    _xmax = max(_x1, _x2)
    _ymin = min(_y1, _y2)
    _ymax = max(_y1, _y2)
    tool.delHandler("motion_notify")
    _active_layer = gtkimage.getActiveLayer()
    _layers = []
    _layers.append(gtkimage.getTopLayer())
    _objdict = {}
    while len(_layers):
        _layer = _layers.pop()
        if _layer is not _active_layer:
            if _layer.getVisibility():
                _objs = _layer.objsInRegion(_xmin, _ymin, _xmax, _ymax)
                if len(_objs):
                    _objdict[_layer] = _objs
        if _layer.hasChildren():
            _layers = _layers + _layer.getChildren()
    if len(_objdict):
        for _layer in _objdict:
            if _layer is not _active_layer:
                for _obj in _objdict[_layer]:
                    gtkimage.transferObject(_obj, _layer, _active_layer)
        gtkimage.redraw()
    tool.clearLocation()
    tool.clearCurrentPoint()
    tool.setHandler("button_press", transfer_object_button_press_cb)
    
def transfer_object_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    _tol = gtkimage.getTolerance()
    _active_layer = gtkimage.getActiveLayer()
    _objdict = gtkimage.mapPoint(_x, _y, _tol, None)
    if len(_objdict):
        for _layer in _objdict:
            if _layer is not _active_layer:
                for _obj, _pt in _objdict[_layer]:
                    gtkimage.transferObject(_obj, _layer, _active_layer)
        gtkimage.redraw()
    else:
        tool.pushObject(_x)
        tool.pushObject(_y)
        tool.setLocation(_x, _y)
        tool.setHandler("motion_notify", select_motion_notify)
        tool.setHandler("button_press", transfer_end_button_press_cb)
        gc = gtkimage.getGC()
        gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                               gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
        gc.set_function(gtk.gdk.INVERT)
        
def transfer_object_init(tool):
    tool.initialize()
    tool.setHandler("button_press", transfer_object_button_press_cb)
    
#
# change color
#

def change_color_init(gtkimage):
    _window = gtkimage.getWindow()
    _dialog = gtk.ColorSelectionDialog("Set Active Color")
    _dialog.set_transient_for(_window)
    _colorsel = _dialog.colorsel

    _prev_color = gtkimage.getOption('LINE_COLOR')
    if gtkimage.hasGTKColor(_prev_color):
        _gtk_color = gtkimage.getGTKColor(_prev_color)
    else:
        _da = gtkimage.getDA()
        _cstring = str(_prev_color)
        _gtk_color = _da.get_colormap().alloc_color(_cstring)
        gtkimage.saveGTKColor(_prev_color, _gtk_color)
    _colorsel.set_previous_color(_gtk_color)
    _colorsel.set_current_color(_gtk_color)
    _colorsel.set_has_palette(gtk.TRUE)
    _response = _dialog.run()
    if _response == gtk.RESPONSE_OK:
        _gtk_color = _colorsel.get_current_color()
        _red = int((_gtk_color.red/65535.0)*255)
        _green = int((_gtk_color.green/65535.0)*255)
        _blue = int((_gtk_color.blue/65535.0)*255)
        _col = color.Color(_red, _green, _blue)
        if gtkimage.hasSelection():
            _active_layer = gtkimage.getActiveLayer()
            for _layer, _obj in gtkimage.getSelectedObjects():
                if _layer is _active_layer:
                    if isinstance(_obj, graphicobject.GraphicObject):
                        _obj.setColor(_col)
    _dialog.destroy()
    gtkimage.redraw()

#
# change linetypes
#

def change_linetype_init(gtkimage):
    window = gtkimage.getWindow()
    dialog = gtk.Dialog("Change Linetype", window,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_OK, gtk.RESPONSE_OK,
                        gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    hbox = gtk.HBox(gtk.FALSE, 2)
    hbox.set_border_width(2)
    dialog.vbox.pack_start(hbox, gtk.FALSE, gtk.FALSE, 0)
    
    label = gtk.Label("Linetype:")
    hbox.pack_start(label, gtk.FALSE, gtk.FALSE, 0)
    cur_linetype = gtkimage.getOption('LINE_TYPE')
    linetypes = gtkimage.getAllLinetypes()
    menu = gtk.Menu()
    idx = 0
    for i in xrange(len(linetypes)):
        lt = linetypes[i]
        if lt is cur_linetype:
            idx = i
        item = gtk.MenuItem(lt.getName())
        menu.append(item)
    opt_menu = gtk.OptionMenu()
    opt_menu.set_menu(menu)
    opt_menu.set_history(idx)
    hbox.pack_start(opt_menu, gtk.TRUE, gtk.TRUE, 0)
    dialog.show_all()
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        lt = linetypes[opt_menu.get_history()]
        if gtkimage.hasSelection():
            active_layer = gtkimage.getActiveLayer()
            for lyr, obj in gtkimage.getSelectedObjects():
                if lyr is active_layer:
                    if isinstance(obj, graphicobject.GraphicObject):
                        obj.setLinetype(lt)
    dialog.destroy()
    gtkimage.redraw()

#
# change thickness
#

def change_thickness_init(gtkimage):
    window = gtkimage.getWindow()
    dialog = gtk.Dialog("Change Thickness", window,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_OK, gtk.RESPONSE_OK,
                        gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    hbox = gtk.HBox(gtk.FALSE, 2)
    hbox.set_border_width(2)
    dialog.vbox.pack_start(hbox, gtk.FALSE, gtk.FALSE, 0)
    label = gtk.Label("Thickness:")
    hbox.pack_start(label, gtk.FALSE, gtk.FALSE, 0)
    thickness = gtkimage.getOption('LINE_THICKNESS')
    adj = gtk.Adjustment(thickness, 0.0001, 20.0, 0.1, 1.0, 1.0)
    sb = gtk.SpinButton(adj)
    sb.set_digits(1)
    sb.set_numeric(gtk.FALSE)
    hbox.pack_start(sb, gtk.TRUE, gtk.TRUE, 0)
    dialog.show_all()
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        new_thickness = float(sb.get_value())
        if gtkimage.hasSelection():
            active_layer = gtkimage.getActiveLayer()
            for lyr, obj in gtkimage.getSelectedObjects():
                if lyr is active_layer:
                    if isinstance(obj, graphicobject.GraphicObject):
                        obj.setThickness(new_thickness)
    dialog.destroy()
    gtkimage.redraw()

#
# change the style
#

def change_style_init(gtkimage):
    window = gtkimage.getWindow()
    dialog = gtk.Dialog("Change Style", window,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_OK, gtk.RESPONSE_OK,
                        gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
    hbox = gtk.HBox(gtk.FALSE, 2)
    hbox.set_border_width(2)
    dialog.vbox.pack_start(hbox, gtk.FALSE, gtk.FALSE, 0)
    label = gtk.Label("Style:")
    cur_style = gtkimage.getOption('LINE_STYLE')
    styles = gtkimage.getAllStyles()
    menu = gtk.Menu()
    idx = 0
    for i in xrange(len(styles)):
        s = styles[i]
        if s is cur_style:
            idx = i
        item = gtk.MenuItem(s.getName())
        menu.append(item)
    opt_menu = gtk.OptionMenu()
    opt_menu.set_menu(menu)
    opt_menu.set_history(idx)
    hbox.pack_start(label, gtk.FALSE, gtk.FALSE, 0)
    hbox.pack_start(opt_menu, gtk.TRUE, gtk.TRUE, 0)
    dialog.show_all()
    response = dialog.run()
    if response == gtk.RESPONSE_OK:
        s = styles[opt_menu.get_history()]
        if gtkimage.hasSelection():
            active_layer = gtkimage.getActiveLayer()
            for lyr, obj in gtkimage.getSelectedObjects():
                if lyr is active_layer:
                    if isinstance(obj, graphicobject.GraphicObject):
                        obj.setStyle(s)
    dialog.destroy()
    gtkimage.redraw()

#
# arbitrary zoom
#

def zoom_end_button_press_cb(gtkimage, widget, event, tool):
    _xp, _yp = gtkimage.getPoint()
    _x1, _y1 = tool.getLocation()
    _xmin = min(_xp, _x1)
    _ymin = min(_yp, _y1)
    _width, _height = gtkimage.getSize()
    _fw = float(_width)
    _fh = float(_height)
    _wpp = abs(_x1 - _xp)/_fw
    _hpp = abs(_y1 - _yp)/_fh
    if _wpp > _hpp:
        _scale = _wpp
    else:
        _scale = _hpp
    gtkimage.setView(_xmin, _ymin, _scale)
    gtkimage.setPrompt("Click in the drawing.")
    gtkimage.redraw()
    zoom_init(tool)

def zoom_motion_notify(gtkimage, widget, event, tool):
    _tx, _ty = tool.getLocation()
    _px, _py = gtkimage.coordToPixTransform(_tx, _ty)
    # width, height = gtkimage.getSize()
    _gc = gtkimage.getGC()
    _x = event.x
    _y = event.y
    _cp = tool.getCurrentPoint()
    #
    # it would be nice to draw the rectangle in the current
    # shape of the window ...
    #
    if _cp is not None:
        _xc, _yc = _cp
        _xmin = min(_px, _xc)
        _ymin = min(_py, _yc)
        _rw = abs(_xc - _px)
        _rh = abs(_yc - _py)
        widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)
    _xmin = min(_x, _px)
    _ymin = min(_y, _py)
    tool.setCurrentPoint(_x, _y)
    _rw = abs(_x - _px)
    _rh = abs(_y - _py)
    widget.window.draw_rectangle(_gc, gtk.FALSE, _xmin, _ymin, _rw, _rh)

def zoom_button_press_cb(gtkimage, widget, event, tool):
    _x, _y = gtkimage.getPoint()
    tool.setLocation(_x, _y)
    tool.setHandler("motion_notify", zoom_motion_notify)
    tool.setHandler("button_press", zoom_end_button_press_cb)
    gtkimage.setPrompt("Click a second point to define the zoom window.")
    _gc = gtkimage.getGC()
    _gc.set_line_attributes(1, gtk.gdk.LINE_SOLID,
                            gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)
    _gc.set_function(gtk.gdk.INVERT)

def zoom_init(tool):
    tool.initialize()
    tool.setHandler("button_press", zoom_button_press_cb)
