#
# Copyright (c) 2002, Art Haas
#
# This file is part of PythonCAD.
# 
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# code for setting the preferences of an image
#

import gtk

import Generic.text as text
import Generic.units as units

def set_current_prefs(gtkimage, prefdict):
    assert prefdict.has_key('THICKNESS_SB'), "Missing THICKNESS_SB"
    gtkimage.setOption('LINE_THICKNESS', prefdict['THICKNESS_SB'].get_value())
    assert prefdict.has_key('FONT_SEL'), "Missing FONT_SEL"
    fn = prefdict['FONT_SEL'].get_font_name()
    family, style, weight, stretch, size = text.parse_font(fn)
    gtkimage.setOption('FONT_FAMILY', family)
    gtkimage.setOption('FONT_SIZE', size)
    gtkimage.setOption('FONT_STYLE', style)
    gtkimage.setOption('FONT_WEIGHT', weight)
    assert prefdict.has_key('CHAMFER_SB'), "Missing CHAMFER_SB"
    gtkimage.setOption('CHAMFER_LENGTH', prefdict['CHAMFER_SB'].get_value())
    assert prefdict.has_key('FILLET_SB'), "Missing FILLET_SB"
    gtkimage.setOption('FILLET_RADIUS', prefdict['FILLET_SB'].get_value())
    assert prefdict.has_key('UNIT_MENU'), "Missing UNIT_MENU"
    gtkimage.setUnits(prefdict['UNIT_MENU'].get_history())
    assert prefdict.has_key('HIGHLIGHT_CB'), "Missing HIGHLIGHT_CB"
    gtkimage.setOption('HIGHLIGHT_POINTS', prefdict['HIGHLIGHT_CB'].get_active())

def add_misc_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)

    cb = gtk.CheckButton("Highlight Points")
    state = gtkimage.getOption('HIGHLIGHT_POINTS')
    cb.set_active(state)
    pd['HIGHLIGHT_CB'] = cb
    vbox.pack_start(cb, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Line Thickness:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    th = gtkimage.getOption('LINE_THICKNESS')
    adj = gtk.Adjustment(th, 0.0, 100.0, 0.1, 1.0, 1.0)
    sb = gtk.SpinButton(adj)
    sb.set_digits(2)
    sb.set_numeric(gtk.FALSE)
    pd['THICKNESS_SB'] = sb
    hbox.pack_start(sb, gtk.TRUE, gtk.TRUE, 0)
    vbox.pack_start(hbox, gtk.FALSE, gtk.FALSE)
    vbox.show_all()
    
    lb = gtk.Label("Miscellaneous")
    nb.append_page(vbox, lb)
    
def add_font_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Font Properties")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)
    
    fw = gtk.FontSelection()
    fontlist = []
    fontlist.append(gtkimage.getOption('FONT_FAMILY'))
    weight = gtkimage.getOption('FONT_WEIGHT')
    if weight != text.NORMAL:
        if weight == 1:
            fontlist.append('Light')
        elif weight == 2:
            fontlist.append('Bold')
        elif weight == 3:
            fontlist.append('Heavy')
    style = gtkimage.getOption('FONT_STYLE')
    if style != text.NORMAL:
        if style == 1:
            fontlist.append('Oblique')
        elif style == 2:
            fontlist.append('Italic')
    fontlist.append(str(gtkimage.getOption('FONT_SIZE')))
    fw.set_font_name(' '.join(fontlist))
    pd['FONT_SEL'] = fw
    frame.add(fw)
    vbox.show_all()

    lb = gtk.Label("Fonts")
    nb.append_page(vbox, lb)

def add_fillets_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Fillets")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)

    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default radius:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    fr = gtkimage.getOption('FILLET_RADIUS')
    adj = gtk.Adjustment(fr, 0.0, 100.0, 0.1, 1.0, 1.0)
    sb = gtk.SpinButton(adj)
    sb.set_digits(2)
    sb.set_numeric(gtk.FALSE)
    pd['FILLET_SB'] = sb
    hbox.pack_start(sb, gtk.TRUE, gtk.TRUE, 0)
    frame.add(hbox)

    vbox.show_all()

    tl = gtk.Label("Fillets")
    nb.append_page(vbox, tl)
    
def add_chamfers_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Chamfers")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    lb = gtk.Label("Default length:")
    hbox.pack_start(lb, gtk.FALSE, gtk.FALSE)
    fs = gtkimage.getOption('CHAMFER_LENGTH')
    adj = gtk.Adjustment(fs, 0.0, 100.0, 0.1, 1.0, 1.0)
    sb = gtk.SpinButton(adj)
    sb.set_digits(2)
    sb.set_numeric(gtk.FALSE)
    pd['CHAMFER_SB'] = sb
    hbox.pack_start(sb, gtk.TRUE, gtk.TRUE, 0)
    frame.add(hbox)

    tl = gtk.Label("Chamfers")
    nb.append_page(vbox, tl)

def add_units_page(gtkimage, pd, nb):
    vbox = gtk.VBox(gtk.FALSE, 2)
    frame = gtk.Frame("Units")
    vbox.pack_start(frame, gtk.FALSE, gtk.FALSE)
    
    hbox = gtk.HBox(gtk.FALSE, 2)
    ml = gtk.Label("Units:")
    hbox.pack_start(ml, gtk.FALSE, gtk.FALSE)
    menu = gtk.Menu()
    idx = 0
    unit_list = units.get_all_units()
    cur_unit = gtkimage.getOption('UNITS')
    for i in xrange(len(unit_list)):
        if i == cur_unit:
            idx = i
        item = gtk.MenuItem(unit_list[i])
        menu.append(item)
    opt_menu = gtk.OptionMenu()
    opt_menu.set_menu(menu)
    opt_menu.set_history(idx)
    pd['UNIT_MENU'] = opt_menu
    hbox.pack_start(opt_menu, gtk.FALSE, gtk.FALSE)
    frame.add(hbox)
    vbox.show_all()
    
    tl = gtk.Label("Units")
    nb.append_page(vbox, tl)

def check_prefs(gtkimage):
    window = gtkimage.getWindow()
    dialog = gtk.Dialog("Preference", window,
                        gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                         gtk.STOCK_APPLY, gtk.RESPONSE_APPLY,
                         gtk.STOCK_OK, gtk.RESPONSE_OK))
    notebook = gtk.Notebook()
    pref_dict = {}
    dialog.vbox.pack_start(notebook, expand=gtk.TRUE)
    add_units_page(gtkimage, pref_dict, notebook)
    add_chamfers_page(gtkimage, pref_dict, notebook)
    add_fillets_page(gtkimage, pref_dict, notebook)
    add_font_page(gtkimage, pref_dict, notebook)
    add_misc_page(gtkimage, pref_dict, notebook)
    dialog.show_all()
    response = dialog.run()
    if response == gtk.RESPONSE_OK or response == gtk.RESPONSE_APPLY:
        set_current_prefs(gtkimage, pref_dict)
        gtkimage.redraw()
    dialog.destroy()
